#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/config/collector/ResourceConfigCollector.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 16.10.2019
# Last Modified Date: 22.04.2020
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from typing import Dict, Any, Optional, Union
from ampel.log.AmpelLogger import AmpelLogger, VERBOSE
from ampel.config.collector.AbsDictConfigCollector import AbsDictConfigCollector


class ResourceConfigCollector(AbsDictConfigCollector):

	def __init__(self,
		conf_section: str, content: Optional[Dict] = None,
		logger: Optional[AmpelLogger] = None, verbose: bool = False
	):
		super().__init__(conf_section, content, logger, verbose)

		# Used to temporarily save distribution/source conf information of aliases
		# (usefuly in case of conflicts)
		self.tmp_resource: Dict[str, Any] = {}


	def add(self,
		arg: Dict[str, Any],
		dist_name: str,
		version: Union[str, float, int],
		register_file: str
	) -> None:
		""" """

		if not isinstance(arg, dict):
			self.error(
				f"Resource value must be a dict. "
				f"Offending value {arg}\n"
				f"{self.distrib_hint(register_file, dist_name)}"
			)
			return

		for k, v in arg.items():

			try:

				key = k

				# Global resource
				if k and k[0] == "%":
					key = k[1:]
					scope = "global"
				else:
					# Distribution scoped alias
					if dist_name:
						key = f"{dist_name}/{k}"
						scope = "scoped"
					else:
						scope = ""

				self.tmp_resource[key] = register_file, dist_name

				if self.verbose:
					self.logger.log(VERBOSE,
						f"Adding {scope} resource '{k}' " +
						f"from file {register_file}" if register_file else ""
					)

				if self.get(key):
					self.duplicated_entry(
						conf_key = key,
						section_detail = f"{scope} resource",
						new_file = register_file,
						new_dist = dist_name,
						prev_file = self.tmp_resource.get(key, "unknown")[0],
						prev_dist = self.tmp_resource.get(key, "unknown")[1]
					)
					continue

				self.__setitem__(key, v)

			except Exception as e:
				self.error(
					f"Error occured while loading resource {k} " +
					self.distrib_hint(dist_name, register_file),
					exc_info=e
				)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : ampel/logging/AmpelLoggingError.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 03.12.2018
# Last Modified Date: 18.01.2019
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

class AmpelLoggingError(Exception):
	pass



import os
import pathlib
from typing import (  # type: ignore[attr-defined]
    Any,
    Dict,
    Type,
    _GenericAlias,
    get_args,
)

import yaml

from ampel.abstract.AbsSecretProvider import AbsSecretProvider
from ampel.secret.NamedSecret import NamedSecret
from ampel.secret.Secret import Secret


class DirSecretProvider(AbsSecretProvider):
    """Load secrets from a directory, e.g. a mounted k8s Secret"""

    def __init__(self, path: str) -> None:
        """ """
        self._dir = pathlib.Path(path)
        if not self._dir.is_dir():
            raise NotADirectoryError(path)
        if not os.access(self._dir, os.R_OK):
            raise PermissionError(path)

    def tell(self, arg: Secret, ValueType: Type) -> bool:
        """
        Potentially update an initialized Secret instance with
        the actual sensitive information associable with it.
        :returns: True if the Secret was told/resolved or False
        if the provided Secret is unknown to this secret provider
        """

        if isinstance(arg, NamedSecret):
            try:
                with (self._dir / arg.label).open() as f:
                    value = yaml.safe_load(f)
            except FileNotFoundError:
                return False

            if isinstance(ValueType, _GenericAlias):
                ValueType = get_args(ValueType)  # type: ignore[assignment]

            if isinstance(value, ValueType):
                arg.set(value)
                return True

        return False



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/ingest/T0Compiler.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 11.05.2021
# Last Modified Date: 25.11.2021
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from typing import Dict, List, Optional, Set, Union, Tuple, Any
from ampel.content.MetaRecord import MetaRecord
from ampel.types import ChannelId, DataPointId
from ampel.content.DataPoint import DataPoint
from ampel.abstract.AbsDocIngester import AbsDocIngester
from ampel.util.collections import try_reduce
from ampel.abstract.AbsCompiler import AbsCompiler
from ampel.enum.MetaActionCode import MetaActionCode


class T0Compiler(AbsCompiler):
	"""
	Compiles updates to t0 documents arising from different t0 directives
	"""

	def __init__(self, **kwargs) -> None:
		super().__init__(**kwargs)

		# We assume only the usage of only one shaper (that is only one tracid)
		self.register: Dict[
			DataPointId,
			Tuple[
				DataPoint,
				Set[ChannelId],
				Optional[int],           # trace id
				Optional[Dict[str, Any]] # meta extra
			]
		] = {}


	# Override
	def add(self, # type: ignore[override]
		dps: List[DataPoint],
		channel: ChannelId,
		trace_id: Optional[int],
		extra: Optional[Dict[str, Any]] = None
	) -> None:

		r = self.register
		for dp in dps:
			dpid = dp['id']
			if dpid in r:
				r[dpid][1].add(channel)
			else:
				r[dpid] = dp, {channel}, trace_id, extra


	# Override
	def commit(self, ingester: AbsDocIngester[DataPoint], now: Union[int, float], **kwargs) -> None:
		"""
		Note that we let the ingester handle 'ts' and 'updated' values
		"""

		for dp, channel_sets, trace_id, extra in self.register.values():

			lchans = list(channel_sets)
			meta: MetaRecord = {'ts': now, 'run': self.run_id}

			if extra:
				meta.update(extra) # type: ignore[typeddict-item]

			meta['activity'] = [
				{
					'action': MetaActionCode.ADD_CHANNEL,
					'channel': try_reduce(lchans)
				}
			]
			meta['traceid'] = {'shaper': trace_id}

			if self._tag or dp.get('tag'):
				dp['tag'] = [*(self._tag or []), *(dp.get('tag') or [])]
				meta['activity'].append(self._ingest_tag_activity) # type: ignore

			if self.origin and 'origin' not in dp:
				dp['origin'] = self.origin

			dp['channel'] = lchans
			if 'meta' in dp and isinstance(dp['meta'], list):
				dp['meta'].append(meta)
			else:
				dp['meta'] = [meta]

			ingester.ingest(dp)

		self.register.clear()



#!/usr/bin/env python

import asyncio
import enum
import logging
import operator
import os
from datetime import datetime, timedelta
from functools import reduce
from typing import (
    Any,
    cast,
    Dict,
    List,
    Literal,
    Optional,
    Set,
    Tuple,
    Union,
    cast,
    TYPE_CHECKING,
)

from bson import json_util, tz_util, ObjectId
from fastapi import FastAPI, Header, HTTPException, Query, Depends
from fastapi.middleware.cors import CORSMiddleware
from fastapi.responses import Response
from prometheus_client.exposition import choose_encoder

from ampel.abstract.AbsProcessController import AbsProcessController
from ampel.config.AmpelConfig import AmpelConfig
from ampel.core.AmpelContext import AmpelContext
from ampel.core.AmpelController import AmpelController
from ampel.core.UnitLoader import UnitLoader
from ampel.secret.DictSecretProvider import DictSecretProvider
from ampel.secret.AmpelVault import AmpelVault
from ampel.log.LogFlag import LogFlag
from ampel.metrics.AmpelDBCollector import AmpelDBCollector
from ampel.metrics.AmpelProcessCollector import AmpelProcessCollector
from ampel.metrics.AmpelMetricsRegistry import AmpelMetricsRegistry
from ampel.model.ProcessModel import ProcessModel
from ampel.model.StrictModel import StrictModel
from ampel.core.AmpelDB import AmpelDB
from ampel.enum.DocumentCode import DocumentCode
from ampel.util.hash import build_unsafe_dict_id

if TYPE_CHECKING:
    from ampel.protocol.LoggerProtocol import LoggerProtocol


class ProcessCollection(StrictModel):
    processes: List[ProcessModel]


class ProcessStatus(StrictModel):
    name: str
    tier: Literal[0, 1, 2, 3, None]
    status: Literal["running", "idle"]


class ProcessStatusCollection(StrictModel):
    processes: List[ProcessStatus]


class TaskDescription(StrictModel):
    id: str
    processes: List[str]


class TaskDescriptionCollection(StrictModel):
    tasks: List[TaskDescription]


app = FastAPI()
origins = [
    "http://localhost",
    "http://localhost:8080",
]

app.add_middleware(
    CORSMiddleware,
    allow_origins=origins,
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)

log = logging.getLogger("ampel.run.server")
context: AmpelContext = None  # type: ignore[assignment]


class task_manager:
    process_name_to_controller_id: Dict[str, str] = {}
    controller_id_to_task: Dict[str, Tuple[AbsProcessController, asyncio.Task]] = {}
    task_to_processes: Dict[asyncio.Task, List[ProcessModel]] = {}

    @classmethod
    def get_task(cls, name: str):
        return cls.controller_id_to_task[cls.process_name_to_controller_id[name]][1]

    @classmethod
    def get_controller(cls, name: str):
        return cls.controller_id_to_task[cls.process_name_to_controller_id[name]][0]

    @classmethod
    async def add_processes(
        cls, processes: List[ProcessModel]
    ) -> TaskDescriptionCollection:
        """
        Add processes to the active set. If a process requests a controller
        config that was already instantiated, that controller will be updated
        with the process definition, replacing any definition that may exist
        under the same name. If the process requests a new controller
        configuration, a new task will be spawned.
        """
        global context
        groups: Dict[str, List[ProcessModel]] = {}
        for pm in processes:
            if not pm.active:
                continue
            controller_id = build_unsafe_dict_id(
                pm.controller.dict(exclude_none=True), ret=str
            )
            if controller_id in groups:
                groups[controller_id].append(pm)
            else:
                groups[controller_id] = [pm]
        for config_id, process_group in groups.items():
            names = [pm.name for pm in process_group]
            # update an existing controller
            if config_id in cls.controller_id_to_task:
                controller, task = cls.controller_id_to_task[config_id]
                # add the controller's existing processes to this group
                for pm in cls.task_to_processes[task]:
                    if pm.name not in names:
                        process_group.append(pm)
                controller.update(context.config, context.loader.vault, process_group)
                # update processes for this task
                cls.task_to_processes[task] = process_group
                # add updated processes to controller mapping
                for name in names:
                    cls.process_name_to_controller_id[name] = config_id
                log.info(
                    f"Updated task {id(task)} ({type(controller).__name__} {[pm.name for pm in process_group]})"
                )
            # spawn a new controller
            else:
                controller = context.loader.new(
                    process_group[0].controller,
                    unit_type=AbsProcessController,
                    config=context.config,
                    vault=context.loader.vault,
                    processes=process_group,
                )
                task = asyncio.create_task(controller.run())
                task.add_done_callback(cls.finalize_task)
                cls.task_to_processes[task] = process_group
                cls.controller_id_to_task[config_id] = (controller, task)
                for pm in process_group:
                    cls.process_name_to_controller_id[pm.name] = config_id
                log.info(
                    f"Launched task {id(task)} ({type(controller).__name__} {[pm.name for pm in process_group]})"
                )
        return TaskDescriptionCollection(
            tasks=[
                TaskDescription(
                    id=config_id,
                    processes=[pm.name for pm in cls.task_to_processes[task]],
                )
                for config_id, (controller, task) in cls.controller_id_to_task.items()
            ]
        )

    @classmethod
    async def remove_processes(cls, names: Set[str]) -> None:
        """
        Remove the named process from the active set.
        """
        global context
        to_remove: Dict[str, List[str]] = {}
        for name in names:
            if (config_id := cls.process_name_to_controller_id.get(name)) is None:
                continue
            if config_id in to_remove:
                to_remove[config_id].append(name)
            else:
                to_remove[config_id] = [name]
        expiring = set()
        for config_id, remove_group in to_remove.items():
            controller, task = cls.controller_id_to_task[config_id]
            keep: List[ProcessModel] = []
            drop: List[ProcessModel] = []
            for pm in cls.task_to_processes[task]:
                [keep, drop][pm.name in remove_group].append(pm)
            if not keep:
                controller.stop()
                # controller is empty; wait for it to exit
                expiring.add(task)
                log.info(
                    f"Stopping task {id(task)} ({type(controller).__name__} (empty))"
                )
            else:
                controller.update(context.config, context.loader.vault, keep)
                # controller still has assigned processes; clean up others
                for pm in drop:
                    cls.process_name_to_controller_id.pop(pm.name)
                cls.task_to_processes[task] = keep
                log.info(
                    f"Removed {[pm.name for pm in drop]} from task {id(task)} ({type(controller).__name__})"
                )
        await asyncio.gather(*expiring)

    @classmethod
    def finalize_task(cls, task: asyncio.Task) -> None:
        log.info(
            f"Task {id(task)} finished ({[pm.name for pm in cls.task_to_processes[task]]})"
        )
        for pm in cls.task_to_processes.pop(task):
            config_id = cls.process_name_to_controller_id.pop(pm.name)
            cls.controller_id_to_task.pop(config_id, None)

    @classmethod
    def get_status(cls, name: str) -> Literal["running", "idle"]:
        if name in cls.process_name_to_controller_id:
            return "running"
        else:
            return "idle"

    @classmethod
    async def shutdown(cls) -> None:
        tasks = cls.task_to_processes.keys()
        # Allow 10 seconds to stop gracefully
        for controller, _ in cls.controller_id_to_task.values():
            controller.stop()
        await asyncio.wait_for(asyncio.gather(*tasks, return_exceptions=True), 10)
        # Cancel any tasks that do not exit in time
        for task in tasks:
            task.cancel()
        await asyncio.gather(*tasks, return_exceptions=True)


@app.on_event("startup")
async def init():
    if type(asyncio.get_event_loop()).__module__ == "uvloop":
        raise RuntimeError(
            "uvloop does not work with OS pipes (https://github.com/MagicStack/uvloop/issues/317). start uvicorn with --loop asyncio."
        )
    global context
    context = AmpelContext.load(
        os.environ.get("AMPEL_CONFIG", "config.yml"),
        vault=AmpelVault(providers=[DictSecretProvider.load(os.environ["AMPEL_SECRETS"])])
        if "AMPEL_SECRETS" in os.environ
        else None,
        freeze_config=False,
    )
    AmpelMetricsRegistry.register_collector(AmpelDBCollector(context.db))
    AmpelMetricsRegistry.register_collector(AmpelProcessCollector(name="server"))


app.on_event("shutdown")(task_manager.shutdown)


@app.post("/config/reload")
async def reload_config() -> TaskDescriptionCollection:
    # NB: async to prevent this running a thread
    config_file = os.environ.get("AMPEL_CONFIG", "config.yml")
    secrets_file = os.environ.get("AMPEL_SECRETS")
    try:
        logging.info(f"Reloading config from {config_file}")
        config = AmpelConfig.load(config_file, freeze=False)
        vault = AmpelVault(providers=[DictSecretProvider.load(secrets_file)]) if secrets_file else None
        db = AmpelDB.new(config, vault)
        loader = UnitLoader(config, db=db, vault=vault)
        # Ensure that process models are valid
        with loader.validate_unit_models():
            processes = AmpelController.get_processes(
                config, raise_exc=True, logger=cast("LoggerProtocol", log)
            )
    except Exception:
        logging.exception(f"Failed to load {config_file}")
        raise HTTPException(
            status_code=500, detail="Failed to reload configuration file"
        )

    # remove processes that are no longer defined or no longer in the active set
    await task_manager.remove_processes(
        set(task_manager.process_name_to_controller_id.keys()).difference(
            [pm.name for pm in processes]
        )
    )

    # update global context
    global context
    context = AmpelContext(config, db, loader)

    # update processes currently in the active set
    await task_manager.add_processes(processes)

    return await get_tasks()


# -------------------------------------
# Metrics
# -------------------------------------


@app.get("/metrics")
async def get_metrics(accept: Optional[str] = Header(None)):
    encoder, content_type = choose_encoder(accept)
    return Response(content=encoder(AmpelMetricsRegistry), media_type=content_type)


# -------------------------------------
# Processes
# -------------------------------------


@app.get("/processes")
async def get_processes(
    tier: Optional[int] = Query(None, ge=0, le=3, description="tier to include"),
    name: Optional[List[str]] = Query(None),
    include: Optional[List[str]] = Query(
        None, description="include processes with names that match"
    ),
    exclude: Optional[List[str]] = Query(
        None, description="exclude processes with names that match"
    ),
    controllers: Optional[List[str]] = Query(
        None, description="include processes with these controllers"
    ),
) -> ProcessCollection:
    processes = AmpelController.get_processes(
        context.config,
        tier=cast(Literal[0, 1, 2, 3], tier),
        match=include,
        exclude=exclude,
        controllers=controllers,
    )
    if name:
        processes = [pm for pm in processes if pm.name in name]
    return ProcessCollection(processes=processes)


@app.get("/processes/status")
async def get_processes_status(
    tier: Optional[int] = Query(None, ge=0, le=3, description="tier to include"),
    name: Optional[List[str]] = Query(None),
    include: Optional[List[str]] = Query(
        None, description="include processes with names that match"
    ),
    exclude: Optional[List[str]] = Query(
        None, description="exclude processes with names that match"
    ),
    controllers: Optional[List[str]] = Query(
        None, description="include processes with these controllers"
    ),
) -> ProcessStatusCollection:
    processes = (
        await get_processes(tier, name, include, exclude, controllers)
    ).processes
    return ProcessStatusCollection(
        processes=[
            ProcessStatus(
                name=pm.name, tier=pm.tier, status=task_manager.get_status(pm.name)
            )
            for pm in processes
        ]
    )


@app.post("/processes/start")
async def start_processes(
    tier: Optional[int] = Query(None, ge=0, le=3, description="tier to include"),
    name: Optional[List[str]] = Query(None),
    include: Optional[List[str]] = Query(
        None, description="include processes with names that match"
    ),
    exclude: Optional[List[str]] = Query(
        None, description="exclude processes with names that match"
    ),
    controllers: Optional[List[str]] = Query(
        None, description="include processes with these controllers"
    ),
) -> TaskDescriptionCollection:
    processes = (
        await get_processes(tier, name, include, exclude, controllers)
    ).processes
    return await task_manager.add_processes(processes)


@app.post("/processes/stop")
async def stop_processes(
    tier: Optional[int] = Query(None, ge=0, le=3, description="tier to include"),
    name: Optional[List[str]] = Query(None),
    include: Optional[List[str]] = Query(
        None, description="include processes with names that match"
    ),
    exclude: Optional[List[str]] = Query(
        None, description="exclude processes with names that match"
    ),
    controllers: Optional[List[str]] = Query(
        None, description="include processes with these controllers"
    ),
) -> TaskDescriptionCollection:
    processes = (
        await get_processes(tier, name, include, exclude, controllers)
    ).processes
    await task_manager.remove_processes({pm.name for pm in processes})
    return await get_tasks()


@app.get("/tasks")
async def get_tasks() -> TaskDescriptionCollection:
    return TaskDescriptionCollection(
        tasks=[
            TaskDescription(
                id=config_id,
                processes=[pm.name for pm in task_manager.task_to_processes[task]],
            )
            for config_id, (
                controller,
                task,
            ) in task_manager.controller_id_to_task.items()
        ]
    )


@app.get("/process/{process}")
async def get_process(process: str) -> ProcessModel:
    for tier in range(4):
        try:
            doc = context.config.get(f"process.t{tier}.{process}", dict, raise_exc=True)
        except Exception:
            continue
        return ProcessModel(**doc)
    else:
        raise HTTPException(status_code=404, detail=f"{process} not found")


@app.post("/process/{process}/start")
async def start_process(process: str) -> TaskDescriptionCollection:
    processes = (
        await get_processes(
            tier=None, name=[process], include=None, exclude=None, controllers=None
        )
    ).processes
    return await task_manager.add_processes(processes)


@app.post("/process/{process}/stop")
async def stop_process(process: str):
    await task_manager.remove_processes({process})


@app.post("/process/{process}/kill")
async def kill_process(process: str):
    try:
        task = task_manager.get_controller(process)
    except KeyError:
        raise HTTPException(status_code=404, detail=f"{process} is not running")
    task.cancel()
    try:
        await task
    except asyncio.CancelledError:
        ...


@app.post("/process/{process}/scale")
async def scale_process(
    process: str, multiplier: int = Query(..., gt=0, description="number of replicas")
) -> None:
    try:
        controller = task_manager.get_controller(process)
    except KeyError:
        raise HTTPException(status_code=404, detail=f"{process} is not running")
    if hasattr(controller, "scale"):
        controller.scale(multiplier=multiplier)  # type: ignore
    else:
        raise HTTPException(
            status_code=405, detail=f"{type(controller).__name__} does not scale"
        )


# -------------------------------------
# DB data
# -------------------------------------


@app.get("/stock/{stock_id}")
def get_stock(stock_id: int):
    doc = json_util._json_convert(
        context.db.get_collection("stock").find_one({"_id": stock_id}),
        json_util.RELAXED_JSON_OPTIONS,
    )
    for k in "created", "updated":
        doc[k] = {facet: datetime.fromtimestamp(ts) for facet, ts in doc[k].items()}
    for jentry in doc["journal"]:
        jentry["ts"] = datetime.fromtimestamp(jentry["ts"])
    return doc


@app.get("/summary/channels")
def stock_summary():
    cursor = context.db.get_collection("stock").aggregate(
        [
            {
                "$facet": {
                    "channels": [
                        {"$unwind": "$channel"},
                        {
                            "$group": {
                                "_id": "$channel",
                                "count": {"$sum": 1},
                            }
                        },
                    ],
                    "total": [{"$count": "count"}],
                }
            }
        ]
    )
    facets = next(cursor)
    return {
        "channels": {
            doc["_id"]: doc["count"]
            for doc in sorted(facets["channels"], key=lambda doc: doc["_id"])
        },
        "total": facets["total"][0]["count"],
    }


def transform_doc(doc: Dict[str, Any], tier: int) -> Dict[str, Any]:
    doc = json_util._json_convert(doc, json_util.RELAXED_JSON_OPTIONS)
    if tier == 1:
        doc["added"] = datetime.fromtimestamp(doc["added"])
    if tier == 2:
        for jentry in doc["journal"]:
            jentry["dt"] = datetime.fromtimestamp(jentry["dt"])
        doc["code"] = DocumentCode(doc["code"]).name
        for subrecord in doc.get("body", []):
            subrecord["ts"] = datetime.fromtimestamp(subrecord["ts"])
    return doc


@app.get("/stock/{stock_id}/t{tier}")
def get_tier_docs(stock_id: int, tier: int = Query(..., ge=0, le=2)):
    cursor = context.db.get_collection(f"t{tier}").find({"stock": stock_id})
    return {
        "matches": [transform_doc(doc, tier) for doc in cursor for doc in cursor],
        "tier": tier,
    }


@app.get("/t2/summary")
def t2_summary():
    cursor = context.db.get_collection("t2").aggregate(
        [
            {
                "$group": {
                    "_id": {"code": "$code", "unit": "$unit"},
                    "count": {"$sum": 1},
                }
            }
        ]
    )
    summary = {}
    for doc in cursor:
        code = DocumentCode(doc["_id"]["code"]).name
        if code not in summary:
            summary[code] = {}
        summary[code][doc["_id"]["unit"]] = doc["count"]
    return summary


# -------------------------------------
# DB var
# -------------------------------------

# abbrevations used by DBLoggingHandler and FilterBlocksHandler
FIELD_ABBREV = {
    "a": "alert",
    "ac": "autocomplete",
    "c": "channel",
    "f": "flags",
    "m": "message",
    "s": "stock",
    "x": "extra",
}


async def query_time(
    after: Optional[Union[timedelta, datetime]] = Query(None),
    before: Optional[Union[timedelta, datetime]] = Query(None),
) -> List[Dict[str, Any]]:
    andlist = []
    if after or before:
        now = datetime.now(tz_util.utc)
        if after:
            andlist.append(
                {
                    "_id": {
                        "$gt": ObjectId.from_datetime(
                            now - after if isinstance(after, timedelta) else after
                        )
                    }
                }
            )
        if before:
            andlist.append(
                {
                    "_id": {
                        "$lt": ObjectId.from_datetime(
                            now - before if isinstance(before, timedelta) else before
                        )
                    }
                }
            )
    return andlist


async def query_event(
    process: Optional[str] = Query(None),
    tier: Optional[int] = Query(None, ge=0, le=3, description="tier to include"),
    time_constraint: List[Dict[str, Any]] = Depends(query_time),
) -> Dict[str, Any]:
    query: Dict[str, Any] = {}
    if process:
        query["process"] = process
    if tier is not None:
        query["tier"] = tier
    if time_constraint:
        query["$and"] = time_constraint
    return query


# NB: handlers that use Mongo are synchronous; FastAPI implicitly runs them
# in a thread. Could also use Motor for this.
@app.get("/events")
@app.get("/events/{process}")
def get_events(base_query: dict = Depends(query_event)):
    cursor = context.db.get_collection("events").find(
        {"run": {"$exists": True}, **base_query}
    )
    return {
        "events": [
            {
                "timestamp": doc["_id"].generation_time,
                **{
                    FIELD_ABBREV.get(k, k): v
                    for k, v in doc.items()
                    if k not in {"_id"}
                },
            }
            for doc in cursor
        ]
    }


@app.get("/logs/{run_id}")
def get_logs(
    run_id: int,
    flags: Optional[  # type: ignore[valid-type]
        List[enum.Enum("LogFlagName", {k: k for k in LogFlag.__members__})]
    ] = Query(None),
):
    query: Dict[str, Any] = {"r": run_id}
    if flags:
        query["f"] = {
            "$bitsAllSet": reduce(
                operator.or_, [LogFlag.__members__[k.name] for k in flags]
            )
        }
    cursor = context.db.get_collection("logs").find(query, {"r": 0})
    translate_keys = {"_id", "f"}
    return {
        "logs": [
            {
                "timestamp": doc["_id"].generation_time,
                "flags": [
                    k for k, v in LogFlag.__members__.items() if (v & doc["f"])
                ],
                **{
                    FIELD_ABBREV.get(k, k): v
                    for k, v in doc.items()
                    if k not in translate_keys
                },
            }
            for doc in cursor
        ]
    }


@app.get("/troubles")
def get_troubles(base_query: dict = Depends(query_event)):
    cursor = context.db.get_collection("troubles").find(base_query)
    return {
        "troubles": [
            {
                "timestamp": doc["_id"].generation_time,
                **{k: v for k, v in doc.items() if k not in {"_id"}},
            }
            for doc in cursor
        ]
    }


if __name__ == "__main__":
    import uvicorn  # type: ignore

    # NB: libuv does not play nice with OS pipes, so concurrent.process will
    # not work with uvloop: https://github.com/MagicStack/uvloop/issues/317
    uvicorn.run(
        "ampel.run.server:app",
        host="127.0.0.1",
        port=5000,
        log_level="info",
        loop="asyncio",
    )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/cli/T2Command.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 16.03.2021
# Last Modified Date: 13.12.2021
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from datetime import datetime
from argparse import ArgumentParser
from typing import Sequence, Dict, Any, Optional, Union
from ampel.core.EventHandler import EventHandler
from ampel.core.AmpelContext import AmpelContext
from ampel.abstract.AbsIdMapper import AbsIdMapper
from ampel.log.AmpelLogger import AmpelLogger
from ampel.log.LogFlag import LogFlag
from ampel.t2.T2Utils import T2Utils
from ampel.util.pretty import prettyjson
from ampel.content.T2Document import T2Document
from ampel.cli.utils import maybe_load_idmapper
from ampel.cli.AbsCoreCommand import AbsCoreCommand
from ampel.cli.ArgParserBuilder import ArgParserBuilder
from ampel.cli.MaybeIntAction import MaybeIntAction
from ampel.cli.LoadJSONAction import LoadJSONAction
from ampel.cli.AmpelArgumentParser import AmpelArgumentParser

hlp = {
	"show": "Show T2 document(s) as JSON (stdout)",
	"save": "Save T2 document(s) as JSON using the provide file path",
	"reset": "Delete body and set the run state to NEW",
	"soft-reset": "Set the run state to NEW",
	"config": "Path to an ampel config file (yaml/json)",
	"secrets": "Path to a YAML secrets store in sops format",
	'out': 'Path to file were output will be written (printed to stdout otherwise)',
	"unit": "Unit id/name",
	"limit": "Limit number of returned documents",
	"unit-config": "Unit config (integer number). Use string 'null' to match null",
	"code": "T2 code (0: COMPLETED, -1: NEW)",
	"link": "T2 document link (hex)",
	"stock": "Stock(s) associated with t2 doc (OR matched if multi-valued)",
	'id-mapper': 'Convert stock ids using the provided id mapper (ex: ZTFIdMapper)',
	"custom-match": "Custom mongodb match as JSON string (ex: {\"body.result.mychi2\": {\"$gt\": 1.0}})",
	'no-resolve-stock': 'Keep stock as int when matching using id-mapper',
	"resolve-config": "Translate 'config' field from int back to dict",
	"human-times": "Translate timestamps to human-readable strings"
}

class T2Command(AbsCoreCommand):

	def __init__(self):
		self.parsers = {}

	# Mandatory implementation
	def get_parser(self, sub_op: Optional[str] = None) -> Union[ArgumentParser, AmpelArgumentParser]:

		if sub_op in self.parsers:
			return self.parsers[sub_op]

		sub_ops = ['show', 'save', 'reset', 'soft-reset']
		if sub_op is None or sub_op not in sub_ops:
			return AmpelArgumentParser.build_choice_help(
				't2', sub_ops, hlp,
				description = 'Allows to match and then view or reset T2 documents.\n' +
				'Reset operations are registered in the event database and a\n' +
				'dedicated journal entry is added to the affected t2 documents.'
			)

		builder = ArgParserBuilder("t2")
		builder.add_parsers(sub_ops, hlp)

		builder.notation_add_note_references()
		builder.notation_add_example_references()

		# Required
		builder.add_arg('required', 'config')
		builder.add_arg('save.required', 'out')

		# Optional general
		builder.add_arg('optional', 'secrets', default=None)
		builder.add_arg('optional', 'id-mapper')
		builder.add_arg('optional', 'debug', action='count', default=0, help='Debug')
		builder.add_arg('optional', 'dry-run', action='store_true')
		builder.add_arg('optional', 'limit', action='store_true')

		# Optional match criteria
		builder.add_group('match', 'Optional T2 documents matching criteria')
		builder.add_arg('match', 'unit', nargs='+', action=MaybeIntAction)
		builder.add_arg('match', 'unit-config', nargs='+', action=MaybeIntAction)
		builder.add_arg('match', 'code', nargs='+', action=MaybeIntAction)
		builder.add_arg('match', 'link', nargs='+')
		builder.add_arg('match', 'stock', nargs='+', action=MaybeIntAction)
		builder.create_logic_args('match', 'run', 'Run id', pos=0, ref='2')
		builder.create_logic_args('match', 'channel', 'Channel', ref='2')
		builder.create_logic_args('match', 'with-tag', 'Tag', ref='2', json=False)
		builder.create_logic_args('match', 'without-tag', 'Tag', excl=True, ref='2', json=False)
		builder.add_arg('match', 'custom-match', metavar='#', action=LoadJSONAction)

		builder.add_group('show|save.format', 'Optional format parameters')
		builder.add_arg('show|save.format', 'resolve-config', action='store_true')
		builder.add_arg('show|save.format', 'human-times', action='store_true')
		builder.add_arg('show|save.format', 'no-resolve-stock', action='store_true')
		

		builder.add_note('reset|soft-reset',
			'Reset operations failing to match any t2 document will not be registered in\n' +
			'in the event database but these will sliently increase the global run counter'
		)

		# Examples
		for el in sub_ops:
			p = f"ampel t2 {el} -config ampel_conf.yaml "
			a = " -out /path/to/file" if el == "save" else ""
			builder.add_example(el, '-unit T2SNCosmo -db.prefix AmpelTest', prepend=p, append=a)
			builder.add_example(el, '-channel MY_CHANNEL -code -1', prepend=p, append=a)
			builder.add_example(el, '-stock 122621027 122620210 -unit DemoTiedLightCurveT2Unit', prepend=p, append=a)

		builder.add_example('show|save', '-db.prefix AmpelTest -human-times -stock ZTF20aaqubac -resolve-config -id-mapper ZTFIdMapper')

		self.parsers = builder.get()
		return self.parsers[sub_op]


	# Mandatory implementation
	def run(self, args: Dict[str, Any], unknown_args: Sequence[str], sub_op: Optional[str] = None) -> None:

		if sub_op is None:
			raise ValueError("A sub-operation (show, save, reset, soft-reset) needs to be specified")

		ctx = self.get_context(args, unknown_args, ContextClass=AmpelContext, one_db=True)
		logger = AmpelLogger.from_profile(
			ctx, 'console_debug' if args['debug'] else 'console_info',
			base_flag=LogFlag.MANUAL_RUN
		)

		args['config'] = args.pop('unit_config')

		t2_utils = T2Utils(logger)
		col = ctx.db.get_collection('t2', mode='r')

		maybe_load_idmapper(args)
		self.convert_logical_args('tag', args)

		# args['id_mapper'] is used for matching whereas id_mapper is potentially discarded for printing
		id_mapper = None if args.get('no_resolve_stock', False) else args['id_mapper']

		if sub_op == 'show':

			m = t2_utils.match_t2s(**args)
			limit = args.get('limit', False)
			if args.get('dry_run'):
				c = col.count_documents(m)
				if limit:
					c = min(limit, c)
				logger.info(
					f"Query: {m}\n"
					f"Number of matched documents: {c}\n"
					f"Exiting (dry-run)"
				)
				return
			else:
				c = col.find(m).limit(limit) if limit else col.find(m)

			if args['resolve_config'] or args['human_times'] or id_mapper:
				resolve_config = args['resolve_config']
				human_times = args['human_times']
				for el in c:
					self.morph_ret(ctx, el, resolve_config, human_times, id_mapper)
					print(prettyjson(el))
			else:
				print(prettyjson(list(c)))

		elif sub_op == 'save':

			c = t2_utils.get_t2s(col=col, **args)
			if args.get('dry_run'):
				logger.info("Exiting (dry-run)")
				return

			resolve_config = args['resolve_config']
			human_times = args['human_times']
			with open(args['out'], 'w') as f:
				f.write("[\n")
				for el in c:
					f.write(
						prettyjson(
							self.morph_ret(ctx, el, resolve_config, human_times, id_mapper)
						) + ",\n"
					)
				f.write("]\n")

		# reset or soft-reset below
		elif sub_op.endswith('reset'):

			run_id = ctx.new_run_id()
			changed = t2_utils \
				.i_know_what_i_am_doing() \
				.reset_t2s(col=col, cli=True, run_id=run_id, soft = (sub_op == 'soft-reset'), **args)

			print('Number of changed T2 docs: %i' % changed)

			if changed > 0:
				# Add new doc in the 'events' collection
				EventHandler('CLI', ctx.db, run_id=run_id, tier=-1)


	def morph_ret(self,
		ctx: AmpelContext, doc: T2Document,
		resolve_config: bool = False,
		human_times: bool = False,
		id_mapper: Optional[AbsIdMapper] = None
	) -> T2Document:

		if resolve_config and doc['config']:
			doc['config'] = ctx.config._config['confid'].get(doc['config'])

		if human_times:
			for el in doc['meta']:
				if 'ts' in el:
					el['ts'] = datetime.utcfromtimestamp(el['ts']).isoformat() # type: ignore

		if id_mapper:
			doc['stock'] = id_mapper.to_ext_id(doc['stock']) # type: ignore[arg-type]

		return doc



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/abstract/AbsProcessorTemplate.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 16.07.2021
# Last Modified Date: 16.07.2021
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from typing import Dict, Any
from ampel.log.AmpelLogger import AmpelLogger
from ampel.base.decorator import abstractmethod
from ampel.model.UnitModel import UnitModel
from ampel.config.builder.AbsConfigTemplate import AbsConfigTemplate


class AbsProcessorTemplate(AbsConfigTemplate, abstract=True):

	@abstractmethod
	def get_model(self, config: Dict[str, Any], logger: AmpelLogger) -> UnitModel:
		...



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/model/purge/PurgeModel.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 18.06.2020
# Last Modified Date: 18.06.2020
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from ampel.model.StrictModel import StrictModel
from ampel.model.purge.PurgeContentModel import PurgeContentModel
from ampel.model.purge.PurgeLogsModel import PurgeLogsModel


class PurgeModel(StrictModel):
	content: PurgeContentModel
	logs: PurgeLogsModel



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/abstract/AbsProcessController.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 06.04.2020
# Last Modified Date: 17.04.2020
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from typing import Dict, Optional, Literal, Sequence, Any, ClassVar
from ampel.base.decorator import abstractmethod
from ampel.base.AmpelABC import AmpelABC
from ampel.base.AmpelBaseModel import AmpelBaseModel
from ampel.config.AmpelConfig import AmpelConfig
from ampel.secret.AmpelVault import AmpelVault
from ampel.model.ProcessModel import ProcessModel
from ampel.log.AmpelLogger import AmpelLogger
from ampel.metrics.AmpelMetricsRegistry import AmpelMetricsRegistry


class AbsProcessController(AmpelABC, AmpelBaseModel, abstract=True):

	config: AmpelConfig
	processes: Sequence[ProcessModel]
	vault: Optional[AmpelVault] = None
	log_profile: str = "default"

	process_count: ClassVar[Any] = AmpelMetricsRegistry.gauge(
		"processes",
		"Number of concurrent processes",
		subsystem=None,
		labelnames=("tier", "process")
	)
	process_exceptions: ClassVar[Any] = AmpelMetricsRegistry.counter(
		"process_exceptions",
		"Number of unhandled exceptions",
		subsystem=None,
		labelnames=("tier", "process")
	)

	@classmethod
	def new(cls,
		tier: Literal[0, 1, 2, 3],
		config_file_path: str,
		match: Optional[Sequence[str]] = None,
		exclude: Optional[Sequence[str]] = None,
		override: Optional[Dict] = None,
		log_profile: str = "default",
		**kwargs
	):
		from ampel.core.AmpelContext import AmpelContext
		if config_file_path:
			context = AmpelContext.load(config_file_path)
		else:
			context = AmpelContext.build(freeze_config=False)

		# Avoid circular imports
		from ampel.core.AmpelController import AmpelController
		proc_models = AmpelController.get_processes(
			context.config, tier=tier, match=match, exclude=exclude, controllers=[cls.__name__],
			logger=AmpelLogger.from_profile(context, log_profile)
		)

		return cls(config=context.config, processes=proc_models, log_profile=log_profile)


	@abstractmethod
	async def run(self) -> Any:
		"""
		Run this controller. This coroutine should not return until all its
		tasks have completed or it receives asyncio.CancelledError.
		"""
		...


	@abstractmethod
	def stop(self, name: Optional[str] = None) -> None:
		"""
		Gracefully stop processes.
		
		:param name: name of process to stop. If None, stop all processes.
		"""
		...

	def update(self,
		config: AmpelConfig,
		vault: Optional[AmpelVault],
		processes: Sequence[ProcessModel],
	) -> None:
		"""Change the configuration of the controller."""
		...



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/cli/DBCommand.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 14.03.2021
# Last Modified Date: 06.10.2021
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from typing import Optional, Sequence, Dict, Any, Union
from argparse import ArgumentParser # type: ignore[import]
from ampel.core.AmpelDB import AmpelDB
from ampel.core.AmpelContext import AmpelContext
from ampel.cli.AbsCoreCommand import AbsCoreCommand
from ampel.cli.ArgParserBuilder import ArgParserBuilder
from ampel.cli.main import AmpelArgumentParser
from ampel.log.AmpelLogger import AmpelLogger
from ampel.log.LogFlag import LogFlag


# Help parameter descriptions
hlp = {
	"import": "Import saved ampel databases",
	"save": "Save a databases into file",
	"delete": "Deletes all database starting with the provided prefix (ex: AmpelTest)",
	"prefix": "Prefix of the collections to delete (ex: AmpelTest)",
	"config": "Path to an ampel config file (yaml/json)",
	"secrets": "Path to a YAML secrets store in sops format",
	"log-profile": "One of: default, compact, headerless, verbose, debug",
	"debug": "debug",
	"force": "Delete potententially existing view before view creation",
}

class DBCommand(AbsCoreCommand):


	def __init__(self):
		self.parsers = {}

	# Mandatory implementation
	def get_parser(self, sub_op: Optional[str] = None) -> Union[ArgumentParser, AmpelArgumentParser]:

		if sub_op in self.parsers:
			return self.parsers[sub_op]

		sub_ops = ["import", "export", "delete", "view"]
		if sub_op is None or sub_op not in sub_ops:
			return AmpelArgumentParser.build_choice_help(
				"db", sub_ops, hlp, description = 'Import, export or delete ampel databases. Create or remove views.'
			)

		builder = ArgParserBuilder("db")
		builder.add_parsers(sub_ops, hlp)
		builder.notation_add_note_references()
		builder.notation_add_example_references()

		# Required
		builder.add_arg("required", "config")
		builder.add_arg("import.required", "in")
		builder.add_arg("export.required", "out")

		builder.add_group('view', 'View arguments')
		builder.add_x_args('view', {'name': 'create', 'action': 'store_true'}, {'name': 'discard', 'action': 'store_true'})
		builder.add_x_args(
			'view.required',
			{'name': 'channel'},
			{'name': 'channels-or', 'nargs': '+'},
			{'name': 'channels-and', 'nargs': '+'}
		)

		# Optional
		builder.add_arg("optional", "secrets")
		builder.add_arg('optional', 'debug', action="store_true")
		builder.add_arg('view.optional', 'force', action="store_true")

		builder.add_example("import", "-in /path/to/file")
		builder.add_example("export", "-out /path/to/file")
		builder.add_example("delete", "-mongo.prefix AmpelTest")
		builder.add_example("view", "-create -channel CHAN1")
		builder.add_example("view", "-create -channels-or CHAN1 CHAN2")
		builder.add_example("view", "-discard -channel CHAN1")

		self.parsers.update(
			builder.get()
		)

		return self.parsers[sub_op]


	# Mandatory implementation
	def run(self, args: Dict[str, Any], unknown_args: Sequence[str], sub_op: Optional[str] = None) -> None:

		if sub_op == "delete":  # cosmetic mainly
			AmpelDB.create_collection = (lambda x: None) # type: ignore

		ctx: AmpelContext = self.get_context(args, unknown_args)
		db = ctx.db

		logger = AmpelLogger.from_profile(
			ctx, 'console_debug' if args['debug'] else 'console_info',
			base_flag = LogFlag.MANUAL_RUN
		)

		if sub_op == "delete":

			logger.info(f"Deleting databases with prefix {ctx.db.prefix}")
			ctx.db.drop_all_databases()
			logger.info("Done")

		elif sub_op == "view":

			if 'create' not in args and 'discard' not in args:
				logger.error("Either provide 'create' or 'discard' in combination with view command")
				return

			try:
				if x := args.get("channel"):
					logger.info(f"{'Creating' if args['create'] else 'Removing'} view for channel {x}")
					db.create_one_view(x, logger, args['force']) if args['create'] else db.delete_one_view(x, logger)
				elif x := args.get("channels_or"):
					logger.info(f"{'Creating' if args['create'] else 'Removing'} view for channels {x}")
					db.create_or_view(x, logger, args['force']) if args['create'] else db.delete_or_view(x, logger)
				elif x := args.get("channels_and"):
					logger.info(f"{'Creating' if args['create'] else 'Removing'} view for channels {x}")
					db.create_and_view(x, logger, args['force']) if args['create'] else db.delete_and_view(x, logger)
				else:
					logger.error("Channel(s) required\n")
					return
			except Exception as e:
				print(e)
				if "already exists" in str(e):
					#logger.info("View already exist")
					logger.info(str(e))
				else:
					raise e

		elif sub_op == "export":
			raise NotImplementedError()

		elif sub_op == "import":
			raise NotImplementedError()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/config/collector/AbsListConfigCollector.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 03.03.2020
# Last Modified Date: 03.03.2020
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from typing import Any, List, Optional, Union
from ampel.base.decorator import abstractmethod
from ampel.base.AmpelABC import AmpelABC
from ampel.config.collector.ConfigCollector import ConfigCollector


class AbsListConfigCollector(ConfigCollector, AmpelABC, abstract=True):

	@abstractmethod
	def add(self,
		arg: List[Any],
		dist_name: str,
		version: Union[str, float, int],
		register_file: str,
	) -> Optional[int]:
		...



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/ingest/T1Compiler.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 01.01.2018
# Last Modified Date: 21.11.2021
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

import xxhash
from struct import pack
from typing import Optional, Dict, List, Union, Tuple, Set, Any, FrozenSet
from ampel.types import ChannelId, DataPointId, StockId, UnitId, UBson
from ampel.content.T1Document import T1Document
from ampel.content.MetaActivity import MetaActivity
from ampel.abstract.AbsDocIngester import AbsDocIngester
from ampel.abstract.AbsCompiler import AbsCompiler, ActivityRegister


class T1Compiler(AbsCompiler):
	"""
	Helps build a minimal set of :class:`T1Document <ampel.content.T1Document.T1Document>`
	that represent a collection of :class:`datapoints <ampel.content.DataPoint.DataPoint>`,
	as viewed through a set of channels.

	Different T1 units may select different subsets of datapoints associated with a stock.
	Only one :class:`~ampel.content.T1Document.T1Document` will be created for each
	effective subselection. This allows downstream calculations that operate on equivalent
	subselections to be performed only once.

	Excluded datapoints may be included as metadata in T1Document.meta.excl
	"""

	#: true: datapoints will have the same hash regardless of their order (unordered set)
	#: false: order of datapoints will influence the computed hash
	sort: bool = True

	#: Convention: negative size means the int hash result will be converted to signed int
	#: (necessary when using mongodb)
	#: Use 64/-64 if you expect more than 10K states per stock(s)
	hash_size: int = -32

	#: Change this only if you know what you're doing
	seed: int = 0


	def __init__(self, **kwargs) -> None:

		super().__init__(**kwargs)
		self.hasher = getattr(xxhash, f'xxh{abs(self.hash_size)}_intdigest')

		# Internal structure used for compiling documents
		self.t1s: Dict[
			Tuple[
				Optional[UnitId],       # unit
				Optional[int],          # config
				StockId,                # stock
				Tuple[DataPointId, ...] # tuple(dps) [will be hashed into link]
			],
			Tuple[
				int, 	                # link
				Set[ChannelId],         # channels (doc)
				UBson,                  # body
				Optional[int],          # code
				Dict[
					FrozenSet[Tuple[str, Any]], # key: traceid
					Tuple[
						ActivityRegister,
						Dict[str, Any]  # meta extra
					]
				],
			]
		] = {}


	def add(self, # type: ignore[override]
		dps: List[DataPointId],
		channel: ChannelId,
		traceid: Dict[str, Any],
		stock: StockId = 0,
		activity: Optional[Union[MetaActivity, List[MetaActivity]]] = None,
		meta_extra: Optional[Dict[str, Any]] = None,
		unit: Optional[UnitId] = None,
		config: Optional[int] = None,
		body: UBson = None,
		code: Optional[int] = None
	) -> int:
		"""
		:param unit: potential AbsT1ComputeUnit subclass to be associated with this doc
		:param config: config of the AbsT1ComputeUnit subclass associated with this doc
		"""

		if not dps:
			raise ValueError("Argument dps cannot be empty")

		if self.sort:
			# Note: int DataPointId is implicitely required,
			# hashing of single dp ids will be required if bytes is used rather than int
			dps = sorted(dps)

		k = (unit, config, stock, tuple(dps))
		tid = frozenset(traceid.items())

		# a: (link, {channels}, body, code, dict[traceid, (ActivityRegister, meta extra)])
		# a: ( 0  ,     1     ,  2  ,  3  ,                   4                        )])
		if a := self.t1s.get(k):

			a[1].add(channel)

			# One meta record will be created for each unique trace id
			# (the present compiler could be used by multiple handlers configured differently)
			# activity can contain channel-based dps exclusions (excl: [1232, 3232])
			if tid in a[4]:

				# (activity register, meta_extra)
				b = a[4][tid]

				# update internal register
				self.register_meta_info(b[0], b[1], channel, activity, meta_extra)
			else:
				a[4][tid] = self.new_meta_info(channel, activity, meta_extra)

		else:

			i = self.hasher(
				pack(f">{len(dps)}q", *dps),
				self.seed
			)

			# Convert unsigned to signed
			if self.hash_size < 0 and i & (1 << (-self.hash_size-1)):
				i -= 2**-self.hash_size

			a = self.t1s[k] = (
				i, {channel}, body, code,
				{tid: self.new_meta_info(channel, activity, meta_extra)}
			)

		return a[0]


	def commit(self, ingester: AbsDocIngester[T1Document], now: Union[int, float], **kwargs) -> None:

		# t1: (unit, config, stock, dps)
		# t2: (link, {channels}, body, code, dict[traceid, (ActivityRegister, meta extra)])
		# t2: ( 0  ,     1     ,  2  ,  3  ,                   4                        )])
		for t1, t2 in self.t1s.items():

			# Note: mongodb maintains key order
			d: T1Document = {'link': t2[0]}

			if t1[0]:
				d['unit'] = t1[0] # type: ignore[typeddict-item]

			if t1[1]:
				d['config'] = t1[1] # type: ignore[typeddict-item]

			d['stock'] = t1[2]

			if self.origin:
				d['origin'] = self.origin

			d['channel'] = list(t2[1])
			d['dps'] = list(t1[3])

			d['meta'], tags = self.build_meta(t2[4], now)

			if tags:
				d['tag'] = tags

			if t2[2]:
				d['body'] = [t2[2]]

			if t2[3] is not None:
				d['code'] = t2[3]

			ingester.ingest(d)

		self.t1s.clear()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/mongo/view/MongoOrView.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 26.03.2021
# Last Modified Date: 21.07.2021
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from typing import Any, Dict
from ampel.mongo.view.AbsMongoFlatMultiView import AbsMongoFlatMultiView

class MongoOrView(AbsMongoFlatMultiView):

	# Override to add other matching criteria (ex: tag-based selection)
	def get_first_match(self) -> Dict[str, Any]:
		return {'$match': {'channel': {'$in': self.channel}}}



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/abstract/AbsT3Selector.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 09.12.2019
# Last Modified Date: 17.02.2020
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from typing import Iterable, ClassVar, Optional
from ampel.base.AmpelABC import AmpelABC
from ampel.base.decorator import abstractmethod
from ampel.core.ContextUnit import ContextUnit


class AbsT3Selector(AmpelABC, ContextUnit, abstract=True):

	#: field used to identify stocks
	field_name: ClassVar[str] = "stock"

	@abstractmethod
	def fetch(self) -> Optional[Iterable]:
		""" Get selected stock ids """



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/dev/DevAmpelContext.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 10.06.2020
# Last Modified Date: 16.11.2021
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from typing import Optional, Any, Dict, Union, List, Type
from importlib import import_module
from ampel.base.AuxUnitRegister import AuxUnitRegister
from ampel.base.LogicalUnit import LogicalUnit
from ampel.core.ContextUnit import ContextUnit
from ampel.core.AmpelDB import AmpelDB
from ampel.core.UnitLoader import UnitLoader
from ampel.core.AmpelContext import AmpelContext
from ampel.config.AmpelConfig import AmpelConfig
from ampel.model.UnitModel import UnitModel
from ampel.model.ChannelModel import ChannelModel
from ampel.log.AmpelLogger import AmpelLogger
from ampel.util.freeze import recursive_unfreeze
from ampel.util.hash import build_unsafe_dict_id
from ampel.util.mappings import set_by_path, dictify
from ampel.util.recursion import walk_and_process_dict


class DevAmpelContext(AmpelContext):


	def __init__(self,
		db_prefix: Optional[str] = None, purge_db: bool = False,
		custom_conf: Optional[Dict[str, Any]] = None, **kwargs
	) -> None:
		"""
		:db_prefix: customizes the db prefix name
		(ex: "AmpelTest" will result in the databases: AmpelTest_data, AmpelTest_var, AmpelTest_ext)
		:purge_db: wheter all databases with the given prefix should be deleted (purged)
		:extra_conf: convenience parameter which allows to overwrite given parameters of the underlying
		ampel config (possibly frozen) dictionnary. Nested dict keys such as 'general.something' are supported.
		"""

		super().__init__(**kwargs)

		if db_prefix:
			dict.__setitem__(self.config._config['mongo'], 'prefix', db_prefix)

		if custom_conf or db_prefix:
			conf = self._get_unprotected_conf()
			for k, v in (custom_conf or {}).items():
				set_by_path(conf, k, v)
			self._set_new_conf(conf)

		if purge_db:
			self.db.drop_all_databases()
			self.db.init_db()


	def add_channel(self, name: Union[int, str], access: List[str] = []):
		cm = ChannelModel(channel=name, access=access, version=0)
		conf = self._get_unprotected_conf()
		for k, v in cm.__dict__.items():
			set_by_path(conf, f"channel.{name}.{k}", v)
		self._set_new_conf(conf)


	def register_units(self, *Classes: Type[LogicalUnit]) -> None:
		for Class in Classes:
			self.register_unit(Class)


	def register_unit(self, Class: Type[LogicalUnit]) -> None:

		dict.__setitem__(
			self.config._config['unit'],
			Class.__name__,
			{
				'fqn': Class.__module__,
				'base': [el.__name__ for el in Class.__mro__[:-1] if 'ampel' in el.__module__],
				'distrib': 'unspecified',
				'file': 'unspecified',
				'version': 'unspecified'
			}
		)

		if issubclass(Class, (LogicalUnit, ContextUnit)):

			if self.loader._dyn_register is None:
				self.loader._dyn_register = {}

			self.loader._dyn_register[Class.__name__] = Class

		else:
			AuxUnitRegister._dyn[Class.__name__] = Class


	def gen_config_id(self, unit: str, arg: Dict[str, Any], logger: Optional[AmpelLogger] = None) -> int:

		if logger is None:
			logger = AmpelLogger.get_logger()

		if self.loader._dyn_register and unit in self.loader._dyn_register:
			Unit = self.loader._dyn_register[unit]
			arg = dictify(Unit(**arg, logger=logger)._trace_content)
		elif fqn := self.config._config['unit'][unit].get('fqn'):
			Unit = getattr(import_module(fqn), fqn.split('.')[-1])
			arg = dictify(Unit(**arg, logger=logger)._trace_content)
		else:
			logger.warn(
				f"Unit {unit} not installed locally. Building *unsafe* conf dict hash: "
				f"changes in unit defaults between releases will go undetected"
			)

		conf_id = build_unsafe_dict_id(arg)

		if conf_id not in self.config._config["confid"]:
			# Works with ReadOnlyDict
			dict.__setitem__(self.config._config["confid"], conf_id, arg)
			(logger or AmpelLogger.get_logger()).info(f"New conf id generated: {conf_id} for {arg}")

		if conf_id not in self.db.conf_ids:
			self.db.add_conf_id(conf_id, arg)

		return conf_id


	def _get_unprotected_conf(self) -> Dict[str, Any]:
		if self.config.is_frozen():
			return recursive_unfreeze(self.config._config) # type: ignore[arg-type]
		return self.config._config


	def _set_new_conf(self, conf: Dict[str, Any]) -> None:
		self.config = AmpelConfig(conf, True)
		self.db = AmpelDB.new(self.config, self.loader.vault, self.db.require_exists, self.db.one_db)
		self.loader = UnitLoader(self.config, db=self.db, vault=self.loader.vault)


	def hash_ingest_directive(self, config: Dict[str, Any], logger: AmpelLogger) -> Dict[str, Any]:

		# Example of conf_dicts keys
		# processor.config.directives.0.combine.0.state_t2.0
 		# processor.config.directives.0.combine.0.state_t2.0.config.t2_dependency.0
 		# processor.config.directives.0.point_t2.0
		conf_dicts: Dict[str, Dict[str, Any]] = {}

		walk_and_process_dict(
			arg = config,
			callback = self._gather_t2_config_callback,
			match = ['point_t2', 'stock_t2', 'state_t2', 't2_dependency'],
			conf_dicts = conf_dicts
		)

		# This does the trick of processing nested config first
		sorted_conf_dicts = {
			k: conf_dicts[k]
			for k in sorted(conf_dicts, key=len, reverse=True)
		}

		for k, d in sorted_conf_dicts.items():

			t2_unit = d["unit"]
			conf = d.get("config", {})

			if not isinstance(conf, dict) or not conf:
				continue

			if logger.verbose > 1:
				logger.debug("Hashing ingest config elements")

			if override := d.get("override"):
				conf = {**conf, **override}

			if logger.verbose > 1:
				logger.debug("Computing hash")

			confid = self.gen_config_id(t2_unit, conf, logger)
			d['config'] = confid

		return config


	def _gather_t2_config_callback(self, path, k, d, **kwargs) -> None:
		""" Used by hash_ingest_config """
		if d[k]:
			for i, el in enumerate(d[k]):
				kwargs['conf_dicts'][f"{path}.{k}.{i}"] = el


	def new_context_unit(self, unit: str, **kwargs) -> ContextUnit:
		"""
		Convenience method (for notebooks mainly).
		Limitation: config of underlying unit cannot contain/define field 'unit'
		"""
		return self.loader.new_context_unit(
			context = self,
			model = UnitModel(unit = unit, config = kwargs)
		)



import sys, importlib
from pip._internal.operations import freeze

if __name__ == "__main__":

	exceptions = {

		# Irrevelant
		"typing-extensions", "prometheus-client",
		"certifi", "six",

		# Global (added for each unit)
		"pydantic", "pymongo", "ujson", "xxhash"
	}

	pip_env = {
		z[0].replace("-", "_"): z[1]
		for el in freeze.freeze()
		if len(z := el.split("==")) > 1
		and z[0] not in exceptions
	}

	start = set(sys.modules.keys())
	importlib.import_module(sys.argv[1])
	diff = set(sys.modules.keys()) - start

	print(
		{
			m: pip_env[m]
			for k in sorted(diff)
			if (m := k.split(".")[0]) in pip_env
		}
	)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/t3/stage/T3BaseStager.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 08.12.2021
# Last Modified Date: 19.12.2021
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from time import time
from typing import Optional, Generator

from ampel.types import ChannelId
from ampel.view.T3Store import T3Store
from ampel.model.UnitModel import UnitModel
from ampel.t3.T3DocBuilder import T3DocBuilder
from ampel.content.T3Document import T3Document
from ampel.abstract.AbsT3Stager import AbsT3Stager
from ampel.abstract.AbsT3ReviewUnit import AbsT3ReviewUnit, T
from ampel.t3.stage.BaseViewGenerator import BaseViewGenerator


class T3BaseStager(AbsT3Stager, T3DocBuilder, abstract=True):
	"""
	Base class for several stagers provided by ampel.
	This class does not implement the method stage(...) required by AbsT3Stager,
	it is up to the subclass to do it according to requirements.
	"""

	# Require single channel for now (T3DocBuilder allows multi-channel)
	channel: Optional[ChannelId] = None


	def get_unit(self, unit_model: UnitModel, chan: Optional[ChannelId] = None) -> AbsT3ReviewUnit:
		return self.context.loader.new_safe_logical_unit(
			unit_model,
			unit_type = AbsT3ReviewUnit,
			logger = self.logger,
			_chan = self.channel or chan
		)


	def proceed(self,
		t3_unit: AbsT3ReviewUnit,
		view_generator: BaseViewGenerator[T],
		t3s: T3Store
	) -> Generator[T3Document, None, None]:
		"""
		Executes the method 'process' of t3 unit with provided views generator and t3 store,
		handle potential t3 unit result
		"""

		ts = time()

		try:
			self.logger.info("Running T3unit", extra={'unit': t3_unit.__class__.__name__})
			if (ret := t3_unit.process(view_generator, t3s)) or self.save_stock_ids:
				if x := self.handle_t3_result(t3_unit, ret, t3s, view_generator.get_stock_ids(), ts):
					yield x

		except Exception as e:
			self.event_hdlr.handle_error(e, self.logger)
		finally:
			self.flush(t3_unit)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : ampel/model/builder/BuilderAliasModel.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 09.10.2019
# Last Modified Date: 11.10.2019
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from typing import Dict, Any, Optional
from ampel.model.StrictModel import StrictModel

class BuilderAliasModel(StrictModel):

	t0: Optional[Dict[str, Any]]
	t1: Optional[Dict[str, Any]]
	t2: Optional[Dict[str, Any]]
	t3: Optional[Dict[str, Any]]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/model/t3/AliasableModel.py
# License           : BSD-3-Clause
# Author            : jvs
# Date              : Unspecified
# Last Modified Date: Unspecified
# Last Modified By  : jvs

from typing import Any, Optional, Type, ClassVar, TYPE_CHECKING
from ampel.model.StrictModel import StrictModel

if TYPE_CHECKING:
	from ampel.config.AmpelConfig import AmpelConfig


class AliasableModel(StrictModel):
	"""
	A model that can be initialized from a global alias in the alias.t3 section
	of an AmpelConfig
	"""

	_config: ClassVar[Optional["AmpelConfig"]] = None

	@classmethod
	def validate(cls: Type["AliasableModel"], value: Any) -> "AliasableModel":
		if cls._config and isinstance(value, str):
			d = cls._config.get(f"alias.t3.%{value}", dict)
			if d:
				value = d
			else:
				raise ValueError(f"{cls.__name__} alias '{value}' not found in Ampel config")
		return super().validate(value)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/abstract/AbsCompiler.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 07.05.2021
# Last Modified Date: 21.11.2021
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from typing import Optional, Literal, Sequence, Union, Dict, FrozenSet, Tuple, Set, Any, List
from ampel.types import Tag, ChannelId
from ampel.base.AmpelABC import AmpelABC
from ampel.base.decorator import abstractmethod
from ampel.base.AmpelBaseModel import AmpelBaseModel
from ampel.abstract.AbsDocIngester import AbsDocIngester
from ampel.content.MetaRecord import MetaRecord
from ampel.content.MetaActivity import MetaActivity
from ampel.enum.MetaActionCode import MetaActionCode


# Type alias
ActivityRegister = Dict[
	FrozenSet[Tuple[str, Any]], # activity stripped out of channel
	Union[
		None, # activity is not channel-bound
		Set[ChannelId] # will be added to MetaActivity during commit
	]
]


class AbsCompiler(AmpelABC, AmpelBaseModel, abstract=True):

	origin: Optional[int] = None
	tier: Literal[-1, 0, 1, 2, 3]
	run_id: int
	tag: Optional[Union[Tag, Sequence[Tag]]]


	def __init__(self, **kwargs) -> None:

		super().__init__(**kwargs)
		self._tag = None
		self._ingest_tag_activity: Optional[MetaActivity] = None

		if self.tag:
			self._ingest_tag_activity = {'action': MetaActionCode.ADD_INGEST_TAG}
			if isinstance(self.tag, (str, int)):
				self._tag = [self.tag]
				self._ingest_tag_activity['tag'] = self.tag
			else:
				self._tag = list(self.tag)
				self._ingest_tag_activity['tag'] = self._tag


	@abstractmethod(var_args=True)
	def add(self) -> None:
		...


	@abstractmethod
	def commit(self, ingester: AbsDocIngester, now: Union[int, float], **kwargs) -> None:
		...


	def register_meta_info(self,
		ar: ActivityRegister,
		extra_register: Dict[str, Any], # meta_extra
		channel: ChannelId,
		activity: Optional[Union[MetaActivity, List[MetaActivity]]] = None,
		meta_extra: Optional[Dict[str, Any]] = None
	) -> None:
		"""
		Note: We could support the type List[Tuple[str, any]] for the parameter activity
		as the dict form is actually superfluous (frozenset(dict.items()) is used in the end)
		but the performance gain is negligible (~80ns) and it complicates static typing
		"""

		add_chan_registered = False

		if activity:

			for el in [activity] if isinstance(activity, dict) else activity:

				# Channel-bound activity
				if 'channel' in el:

					if el['action'] & MetaActionCode.ADD_CHANNEL:
						add_chan_registered = True

					# Strip out 'channel' from activity
					# (to be able to merge similar activities accross channels)
					x = frozenset([y for y in el.items() if y[0] != 'channel'])

					# activity is already associated with another channel
					if x in ar:

						# Should not happen unless ingestion handler has an erroneous behavior
						if (g := ar[x]) is None:
							# Raising an error is probably a bad idea actually
							raise ValueError("Channel-less / channel-bound activity conflict [0]")

						else:
							# Add current channel to previously registered activity
							if isinstance(el['channel'], (int, str)):
								g.add(el['channel'])
							else:
								g.update(el['channel'])
					else:
						ar[x] = {el['channel']} if isinstance(el['channel'], (int, str)) else set(el['channel'])

				# Channel-less activity
				else:

					# Build register key
					x = frozenset(el.items())

					# activity was already registered
					if x in ar:

						# Previously registered activity is associated with a channel
						if ar[x] is not None:
							# Raising an error is probably a bad idea actually
							raise ValueError("Channel-less / channel-bound activity conflict [1]")

					# register new channel-less activity
					else:
						# (dict key contains activity content, value is None as in channel-less)
						ar[x] = None

		if not add_chan_registered:

			# "Add chan" activity 'key' (used in activity dict below)
			add_chan = frozenset(
				[('action', MetaActionCode.ADD_CHANNEL | MetaActionCode.BUMP_STOCK_UPD)]
			)

			if z := ar.get(add_chan):
				z.add(channel)
			else:
				ar[add_chan] = {channel}

		if meta_extra:
			# No collision detection implemented yet
			extra_register.update(meta_extra)


	@staticmethod
	def _metactivity_key(activity: MetaActivity, skip_keys: Optional[set[str]] = None) -> FrozenSet[tuple[str, Any]]:
		return frozenset(
			(
				k,
				tuple(v) if isinstance(v, list) else v
			) for k, v in activity.items()
			if skip_keys is None or k not in skip_keys
		)


	def new_meta_info(self,
		channel: ChannelId,
		activity: Optional[Union[MetaActivity, List[MetaActivity]]] = None,
		meta_extra: Optional[Dict[str, Any]] = None
	) -> Tuple[ActivityRegister, Dict[str, Any]]: # activity register, meta_extra

		ar: ActivityRegister = {}
		add_chan_registered = False

		if activity:
			for el in [activity] if isinstance(activity, dict) else activity:
				if 'tag' in el and isinstance(el['tag'], list):
					el['tag'] = frozenset(el['tag']) # type: ignore
				if 'channel' in el: # Channel-bound activity
					if el['action'] & MetaActionCode.ADD_CHANNEL:
						add_chan_registered = True
					ar[self._metactivity_key(el, {'channel'})] = (
						{el['channel']} if isinstance(el['channel'], (int, str))
						else set(el['channel'])
					)
				else: # Channel-less activity
					ar[self._metactivity_key(el)] = None

		if not add_chan_registered:
			ar[
				frozenset(
					[('action', MetaActionCode.ADD_CHANNEL | MetaActionCode.BUMP_STOCK_UPD)]
				)
			] = {channel}

		return ar, meta_extra.copy() if meta_extra else {}


	def build_meta(self,
		d: Dict[
			FrozenSet[Tuple[str, Any]], # key: traceid
			Tuple[ActivityRegister, Dict[str, Any]] # activity register, meta_extra
		],
		now: Union[int, float]
	) -> Tuple[List[MetaRecord], List[Tag]]:

		recs: List[MetaRecord] = []
		tags = set(self._tag) if self._tag else set()

		# v[1]: dict[traceid, (activity register, meta_extra)]
		for tid, t in d.items():

			mrec: MetaRecord = {
				'run': self.run_id,
				'ts': now,
				'tier': self.tier
			}

			# add meta extra (for example: alert: 12345678)
			mrec.update(t[1]) # type: ignore[typeddict-item]
			al = []

			for activity, chan in t[0].items():

				ad = dict(
					sorted(
						(
							list(activity) +
							[("channel", next(iter(chan)) if len(chan) == 1 else list(chan))]
						)
						if chan else activity
					)
				)

				if 'tag' in ad:
					if isinstance(ad['tag'], frozenset):
						ad['tag'] = list(ad['tag'])
						tags.update(ad['tag'])
					else:
						tags.add(ad['tag'])

				al.append(ad)

			# report ingest tag activity (only to first metarecord)
			if len(recs) == 0 and self._tag:
				al.append(self._ingest_tag_activity) # type: ignore[arg-type]

			mrec['activity'] = al # type: ignore[typeddict-item]
			mrec['traceid'] = dict(sorted(tid))
			recs.append(mrec)

		return recs, list(tags)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/model/time/UnixTimeModel.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 29.09.2018
# Last Modified Date: 06.06.2020
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from typing import Literal
from pydantic import BaseModel


class UnixTimeModel(BaseModel):

	match_type: Literal['unix_time']
	value: int

	def get_timestamp(self, **kwargs) -> int:
		return self.value



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/demo/DemoProcessor.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 14.03.2021
# Last Modified Date: 19.03.2021
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from ampel.abstract.AbsEventUnit import AbsEventUnit
from ampel.model.UnitModel import UnitModel
from ampel.log import AmpelLogger
from ampel.abstract.AbsT3ReviewUnit import AbsT3ReviewUnit


class DemoProcessor(AbsEventUnit):

	parameter_a: int
	parameter_b: int = 200

	def run(self) -> None:

		from random import randint
		import time
		self.parameter_b = randint(3, 10)
		time.sleep(self.parameter_b)

		# Processor/Admin unit must instantiate their own loggger
		logger = AmpelLogger.get_logger()

		# Feedback
		logger.info("Executing run()")

		# Processor units have access to the associated process id
		logger.info(f"Requested by process {self.process_name}")

		# Config parameters
		logger.info(f"Mandatory config parameter 'parameter_a' value: {self.parameter_a}")
		logger.info(f"Optional config parameter 'parameter_b' value: {self.parameter_b}")


		# Process/Admin units have access to AmpelContext
		ctx = self.context

		# Which contains an instance of UnitLoader
		loader = ctx.loader

		# With which base units can be instantiated
		unit = loader.new_logical_unit(
			model = UnitModel(unit = "DemoReviewT3Unit"),
			logger = logger,
			sub_type = AbsT3ReviewUnit
		)

		print(unit)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/cli/JobCommand.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 15.03.2021
# Last Modified Date: 17.10.2021
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

import tarfile
import tempfile
import ujson
import yaml, os, signal, sys
from time import time
from multiprocessing import Queue, Process
from pydantic import BaseModel
from argparse import ArgumentParser
from importlib import import_module
from typing import List, Sequence, Dict, Any, Optional, Union
from urllib.request import urlretrieve
from ampel.abstract.AbsEventUnit import AbsEventUnit
from ampel.abstract.AbsProcessorTemplate import AbsProcessorTemplate
from ampel.model.UnitModel import UnitModel
from ampel.dev.DevAmpelContext import DevAmpelContext
from ampel.log.AmpelLogger import AmpelLogger
from ampel.log.LogFlag import LogFlag
from ampel.util.freeze import recursive_freeze
from ampel.util.mappings import get_by_path
from ampel.cli.AbsCoreCommand import AbsCoreCommand
from ampel.cli.MaybeIntAction import MaybeIntAction
from ampel.cli.AmpelArgumentParser import AmpelArgumentParser
from ampel.model.job.JobModel import InputArtifact, JobModel, ChannelModel, TaskUnitModel, TemplateUnitModel

try:
	import matplotlib as mpl
	mpl.use('Agg')
except Exception:
	pass


class JobCommand(AbsCoreCommand):
	"""
	Processes job definitions (yaml files).
	A job:
	- can be seen as a sharable analysis shema
	- contains defintions of processor unit and config to be run sequentially
	- should contain everything that is required to run the underlying processor definitions,
	  that is: channels or aliases definitions as well as custom resources potentially required by processors.
	- requires access to an ampel config file since the database, unit or resource definitions
	  contained in the config are necessary to run the job.
	"""

	def __init__(self):
		self.parser = None

	# Mandatory implementation
	def get_parser(self, sub_op: Optional[str] = None) -> Union[ArgumentParser, AmpelArgumentParser]:

		if self.parser:
			return self.parser

		parser = AmpelArgumentParser("job")
		parser.set_help_descr({
			"debug": "Debug",
			#"verbose": "increases verbosity",
			"config": "path to an ampel config file (yaml/json)",
			"schema": "path to YAML job file",
			"secrets": "path to a YAML secrets store in sops format",
			"keep-db": "do not reset databases even if so requested by job file",
			"reset-db": "reset databases even if not requested by job file",
			"interactive": "you'll be asked for each task whether it should be run or skipped\n" + \
				"and if applicable if the db should be reset",
			"with-task": "only perform the task with provide indexes (which starts at 0)",
		})

		# Required
		parser.add_arg("config", "required", type=str)
		parser.add_arg("schema", "required")
		parser.add_arg("with-task", "optional", action=MaybeIntAction, nargs='+')
		parser.add_arg("interactive", "optional", action="store_true")
		parser.add_arg("debug", "optional", action="store_true")
		parser.add_arg("keep-db", "optional", action="store_true")
		parser.add_arg("reset-db", "optional", action="store_true")

		# Optional
		parser.add_arg("secrets", type=str)

		# Example
		parser.add_example("job -config ampel_conf.yaml schema job_file.yaml")
		parser.add_example("job -config ampel_conf.yaml -schema job_file.yaml -keep-db -with-task 2")
		return parser

	def _fetch_inputs(self, job: JobModel, task: Union[TaskUnitModel, TemplateUnitModel], logger: AmpelLogger):
		"""
		Ensure that input artifacts exist
		"""
		for artifact in task.inputs.artifacts:
			resolved_artifact = InputArtifact(**job.resolve_expressions(ujson.loads(artifact.json()), task))
			if resolved_artifact.path.exists():
				logger.info(f"Artifact {resolved_artifact.name} exists at {resolved_artifact.path}")
			else:
				logger.info(f"Fetching artifact {resolved_artifact.name} from {resolved_artifact.http.url} to {resolved_artifact.path}")
				os.makedirs(resolved_artifact.path.parent, exist_ok=True)
				with tempfile.NamedTemporaryFile(delete=False) as tf:
					urlretrieve(resolved_artifact.http.url, tf.name)
					try:
						with tarfile.open(tf.name) as archive:
							logger.info(f"{resolved_artifact.name} is a tarball; extracting")
							os.makedirs(resolved_artifact.path)
							archive.extractall(resolved_artifact.path)
						os.unlink(tf.name)
					except tarfile.ReadError:
						os.rename(tf.name, resolved_artifact.path)

	# Mandatory implementation
	def run(self, args: Dict[str, Any], unknown_args: Sequence[str], sub_op: Optional[str] = None) -> None:

		start_time = time()
		logger = AmpelLogger.get_logger(base_flag=LogFlag.MANUAL_RUN)

		if not os.path.exists(args['schema']):
			raise FileNotFoundError(f"Job file not found: '{args['schema']}'")

		tds: List[Dict[str, Any]] = []

		if isinstance(args['with_task'], int):
			args['with_task'] = [args['with_task']]

		with open(args['schema'], "r") as f:
			job = JobModel(**yaml.safe_load(f))

		s = f"Running job {job.name}"
		logger.info(s)

		print(" " + "-"*len(s))

		if job.requirements:
			# TODO: check job repo requirements
			pass

		purge_db = job.mongo.reset or args['reset_db']

		if purge_db and args['keep_db']:
			logger.info("Keeping existing databases ('-keep-db')")
			purge_db = False

		if args['with_task'] and sum(args['with_task']) > 0 and purge_db and not args['reset_db']:
			logger.info("Ampel job file requires db reset but with-task argument was provided")
			logger.info("Please add argument -reset-db to confirm you are absolutely sure...")
			return

		if args['interactive']:

			from ampel.util.getch import yes_no

			try:
				if purge_db:
					purge_db = yes_no("Delete existing databases")

				args['with_task'] = []
				for i, td in enumerate(job.task):
					s = f" [{td.title}]" if td.title else ""
					if yes_no(f"Process task #{i}" + s):
						args['with_task'].append(i)
			except KeyboardInterrupt:
				sys.exit()

		# DevAmpelContext hashes automatically confid from potential IngestDirectives
		ctx = self.get_context(
			args, unknown_args, logger,
			freeze_config = False,
			ContextClass = DevAmpelContext,
			purge_db = purge_db,
			db_prefix = job.mongo.prefix,
			require_existing_db = False,
			one_db = True
		)

		config_dict = ctx.config._config

		# Add channel(s)
		for c in job.channel:
			chan = ChannelModel(**c)
			logger.info(f"Registering job channel '{chan.channel}'")
			dict.__setitem__(config_dict['channel'], chan.channel, c)

		# Add aliase(s)
		for k, v in job.alias.items():
			if 'alias' not in config_dict:
				dict.__setitem__(config_dict, 'alias', {})
			for kk, vv in v.items():
				logger.info(f"Registering job alias '{kk}'")
				if k not in config_dict['alias']:
					dict.__setitem__(config_dict['alias'], k, {})
				dict.__setitem__(config_dict['alias'][k], kk, vv)

		for i, model in enumerate(job.task):

			if isinstance(model, TemplateUnitModel):

				if model.template not in ctx.config._config['template']:
					raise ValueError(f"Unknown process template: {model.template}")

				fqn = ctx.config._config['template'][model.template]
				class_name = fqn.split(".")[-1]
				Tpl = getattr(import_module(fqn), class_name)
				if not issubclass(Tpl, AbsProcessorTemplate):
					raise ValueError(f"Unexpected template type: {Tpl}")

				tpl = Tpl(**model.config)
				morphed_um = tpl \
					.get_model(ctx.config._config, model.dict()) \
					.dict() | {'title': model.title, 'multiplier': model.multiplier}

				if args.get('debug'):
					from ampel.util.pretty import prettyjson
					logger.info("Task model morphed by template:")
					for el in prettyjson(morphed_um, indent=4).split('\n'):
						logger.info(el)

				tds.append(morphed_um)

			else:
				tds.append(model.dict(exclude={"inputs", "outputs"}))

			logger.info(f"Registering job task#{i} with {tds[-1]['multiplier']}x multiplier")

		ctx.config._config = recursive_freeze(config_dict)

		for i, task_dict in enumerate(tds):

			process_name = f"{job.name}#{i}"

			if 'title' in task_dict:
				self.print_chapter(task_dict['title'] if task_dict.get('title') else f"Task #{i}", logger)
				#process_name += f" [{task_dict['title']}]"
				del task_dict['title']
			elif i != 0:
				self.print_chapter(f"Task #{i}", logger)

			if args['with_task'] is not None and i not in args['with_task']:
				logger.info(f"Skipping task #{i} as requested")
				continue

			multiplier = task_dict.pop('multiplier')
			task_dict['override'] = (task_dict.pop('override') or {}) | {'raise_exc': True}

			# Beacons have no real use in jobs (unlike prod)
			if task_dict['unit'] == 'T2Worker' and 'send_beacon' not in task_dict['config']:
				task_dict['config']['send_beacon'] = False

			self._fetch_inputs(job, job.task[i], logger)

			if multiplier > 1:

				ps = []
				qs = []

				signal.signal(signal.SIGINT, signal_handler)
				signal.signal(signal.SIGTERM, signal_handler)

				try:
					for i in range(multiplier):
						q: Queue = Queue()
						p = Process(
							target = run_mp_process,
							args = (q, config_dict, task_dict, process_name),
							daemon = True,
						)
						p.start()
						ps.append(p)
						qs.append(q)

					for i in range(multiplier):
						ps[i].join()
						if (m := qs[i].get()):
							logger.info(f"{task_dict['unit']}#{i} return value: {m}")
				except KeyboardInterrupt:
					sys.exit(1)

			else:

				proc = ctx.loader.new_context_unit(
					model = UnitModel(**job.resolve_expressions(task_dict, job.task[i])),
					context = ctx,
					process_name = process_name,
					sub_type = AbsEventUnit,
					base_log_flag = LogFlag.MANUAL_RUN
				)
				x = proc.run()
				logger.info(f"{task_dict['unit']} return value: {x}")

		dm = divmod(time() - start_time, 60)
		logger.info("Job processing done. Time required: %s minutes %s seconds\n" % (round(dm[0]), round(dm[1])))


	def print_chapter(self, msg: str, logger: AmpelLogger) -> None:
		logger.info(" ")
		logger.info("=" * (space := (len(msg) + 4)))
		logger.info("‖ " + msg + " ‖") # type: ignore
		logger.info("=" * space)
		logger.info(" ")


def run_mp_process(
	queue: Queue,
	config: Dict[str, Any],
	tast_unit_model: Dict[str, Any],
	process_name: str,
	log_profile: str = "default"
) -> None:

	try:

		# Create new context with serialized config
		context = DevAmpelContext.load(config, one_db=True)

		processor = context.loader.new_context_unit(
			model = UnitModel(**tast_unit_model),
			context = context,
			sub_type = AbsEventUnit,
			log_profile = log_profile,
			process_name = process_name
		)

		queue.put(
			processor.run()
		)

	except Exception as e:
		import traceback
		se = str(e)
		queue.put(
			"\n" + "#"*len(se) + "\n" + str(e) + "\n" + "#"*len(se) + "\n" +
			''.join(traceback.format_exception(type(e), e, e.__traceback__))
		)


def signal_handler(sig, frame):
	#import traceback
	print('Interrupt detected')
	#print("Stack frames:")
	#traceback.print_stack(frame)
	raise KeyboardInterrupt()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/ingest/T2Compiler.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 01.01.2018
# Last Modified Date: 21.11.2021
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from typing import Optional, Dict, Union, Tuple, Set, Any, List, FrozenSet
from ampel.types import ChannelId, UnitId, T2Link, StockId
from ampel.content.T2Document import T2Document
from ampel.content.MetaActivity import MetaActivity
from ampel.abstract.AbsDocIngester import AbsDocIngester
from ampel.abstract.AbsCompiler import AbsCompiler, ActivityRegister


class T2Compiler(AbsCompiler):
	"""
	Helps build a minimal set of :class:`T2Document <ampel.content.T2Document.T2Document>`
	Multiple T2 documents associated with the same stock record accross different channels
	are merged into one single T2 document that references all corresponding channels.
	"""

	def __init__(self, col: Optional[str] = None, **kwargs) -> None:

		super().__init__(**kwargs)
		self.col = col
		self.t2s: Dict[
			# key: (unit name, unit config, link, stock)
			Tuple[UnitId, Optional[int], T2Link, StockId],
			Tuple[
				Set[ChannelId], # channels (doc)
				Dict[
					FrozenSet[Tuple[str, Any]], # key: traceid
					Tuple[ActivityRegister, Dict[str, Any]] # activity register, meta_extra
				]
			]
		] = {}


	def add(self, # type: ignore[override]
		unit: UnitId,
		config: Optional[int],
		stock: StockId,
		link: T2Link,
		channel: ChannelId,
		traceid: Dict[str, Any],
		activity: Optional[Union[MetaActivity, List[MetaActivity]]] = None,
		meta_extra: Optional[Dict[str, Any]] = None
	) -> None:
		"""
		:param tag: tag(s) to be added to T2Document. A corresponding dedicated channel-less
		meta entry will be created (with actin MetaActionCode.ADD_INGEST_TAG and/or ADD_OTHER_TAG)
		"""

		# Doc id
		k = (unit, config, link, stock)
		tid = frozenset(traceid.items())

		# a: tuple({channels}, dict[traceid, (activity register, meta_extra)])
		if a := self.t2s.get(k):

			a[0].add(channel) # set of all channels

			# One meta record will be created for each unique trace id
			# (the present compiler could be used by multiple handlers configured differently)
			if tid in a[1]:

				# (activity register, meta_extra)
				b = a[1][tid]

				# update internal register
				self.register_meta_info(b[0], b[1], channel, activity, meta_extra)

			else:
				a[1][tid] = self.new_meta_info(channel, activity, meta_extra)
		else:
			self.t2s[k] = {channel}, {tid: self.new_meta_info(channel, activity, meta_extra)}


	def commit(self, ingester: AbsDocIngester[T2Document], now: Union[int, float], **kwargs) -> None:

		for k, v in self.t2s.items():

			# Note: mongodb maintains key order
			d: T2Document = {'unit': k[0], 'config': k[1], 'link': k[2]}

			if self.col:
				d['col'] = self.col

			d['stock'] = k[3]

			if self.origin:
				d['origin'] = self.origin

			d['channel'] = list(v[0])
			d['meta'], tags = self.build_meta(v[1], now)

			if tags:
				d['tag'] = tags

			ingester.ingest(d)

		self.t2s.clear()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/model/ingest/T1Combine.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 10.03.2020
# Last Modified Date: 27.05.2021
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from typing import Optional, Union, List, Sequence
from ampel.model.UnitModel import UnitModel
from ampel.model.ingest.T2Compute import T2Compute

class T1Combine(UnitModel):
	"""
	Combine datapoints into t1 document and create associated t2 documents
	"""

	#: Filter result codes that should trigger this T1.
	#: If not specified, T1 documents will be created in response to any passing alert.
	group: Optional[Union[int, List[int]]]

	#: Create or update :class:`T2 documents <ampel.content.T2Document.T2Document>`
	#: bound to :class:`compounds <ampel.content.T1Document.T1Document>`
	state_t2: Optional[Sequence[T2Compute]] = []

	#: Create or update :class:`T2 documents <ampel.content.T2Document.T2Document>`
	#: bound to :class:`datapoints <ampel.content.DataPoint.DataPoint>`
	#: based on the t1 combine unit result
	point_t2: Optional[Sequence[T2Compute]]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/aux/filter/AbsLogicOperatorFilter.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 14.01.2020
# Last Modified Date: 18.06.2020
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from typing import Union, Generic, List, Sequence, Any
from ampel.types import T
from ampel.base.decorator import abstractmethod
from ampel.model.operator.AllOf import AllOf
from ampel.model.operator.FlatAnyOf import FlatAnyOf
from ampel.model.aux.FilterCriterion import FilterCriterion
from ampel.abstract.AbsApplicable import AbsApplicable


class AbsLogicOperatorFilter(Generic[T], AbsApplicable, abstract=True):

	filters: Union[
		FilterCriterion,
		FlatAnyOf[FilterCriterion],
		AllOf[FilterCriterion]
	]

	@staticmethod
	@abstractmethod
	def _apply_filter(args: Sequence[T], f: FilterCriterion) -> List[T]:
		...

	def apply(self, args: Sequence[T]) -> List[T]:

		if isinstance(self.filters, FilterCriterion):
			return self._apply_filter(args, self.filters)

		if isinstance(self.filters, AllOf):
			ret: Any = args
			for f in self.filters.all_of:
				ret = self._apply_filter(ret, f)
			return ret

		if isinstance(self.filters, FlatAnyOf):
			ret = []
			for f in self.filters.any_of:
				for el in self._apply_filter(args, f):
					if el not in ret:
						ret.append(el)
			return ret

		raise ValueError("Incorrect type provided for parameter 'filters'")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/config/collector/AbsForwardConfigCollector.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 16.10.2019
# Last Modified Date: 02.03.2020
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from typing import List, Union, Optional, Sequence, Dict, Any, Type
import ampel.config.builder.FirstPassConfig as fpc # avoid circular import issue
from ampel.base.decorator import abstractmethod
from ampel.base.AmpelABC import AmpelABC
from ampel.log.AmpelLogger import AmpelLogger, VERBOSE


class AbsForwardConfigCollector(dict, AmpelABC, abstract=True):


	def __init__(self,
		# Forward reference type hint to avoid cyclic import issues
		root_config: 'fpc.FirstPassConfig',
		conf_section: str,
		target_collector_type: Type,
		logger: Optional[AmpelLogger] = None,
		verbose: bool = False,
	) -> None:

		self.has_error = False
		self.verbose = verbose
		self.root_config = root_config
		self.conf_section = conf_section
		self.target_collector_type = target_collector_type
		self.logger = AmpelLogger.get_logger() if logger is None else logger

		if verbose:
			self.logger.log(VERBOSE,
				f'Creating {self.__class__.__name__} collector '
				f'for config section "{conf_section}"'
			)


	def add(self,
		arg: Union[Dict[str, Any], List[Any], str],
		dist_name: str,
		version: Union[str, float, int],
		register_file: str,
	) -> None:


		for el in [arg] if isinstance(arg, (dict, str)) else arg:

			path_elements = self.get_path(el, dist_name, version, register_file)
			if not path_elements:
				self.error(f' Follow-up error: could not identify routing for {el}')
				continue

			d = self.root_config
			for path in path_elements:
				d = d[path]

			if not isinstance(d, self.target_collector_type):
				self.error(
					f'Routing destination must be an instance of {self.target_collector_type}'
				)
				self.error(
					f'Type of config element with path '
					f'{".".join(str(x) for x in path_elements)}: {type(d)}'
				)
				return

			d.add(el, dist_name, version, register_file) # type: ignore


	@abstractmethod
	def get_path(self,
		arg: Union[Dict[str, Any], str],
		dist_name: str,
		version: Union[str, float, int],
		register_file: str,
	) -> Optional[Sequence[Union[int, str]]]:
		...


	def error(self, msg: str, exc_info: Optional[Any] = None) -> None:
		self.logger.error(msg, exc_info=exc_info)
		self.has_error = True



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/test/test_server.py
# License           : BSD-3-Clause
# Author            : jvs
# Date              : Unspecified
# Last Modified Date: 11.02.2021
# Last Modified By  : jvs

import asyncio, pytest, yaml
from datetime import datetime
from io import StringIO
from httpx import AsyncClient
from prometheus_client.parser import text_fd_to_metric_families

from ampel.metrics.AmpelDBCollector import AmpelDBCollector
from ampel.metrics.AmpelMetricsRegistry import AmpelMetricsRegistry
from ampel.run import server
from ampel.enum.DocumentCode import DocumentCode
from ampel.util.freeze import recursive_unfreeze
from ampel.util.mappings import set_by_path


@pytest.fixture
async def test_client(dev_context, monkeypatch):
    monkeypatch.setattr("ampel.run.server.context", dev_context)
    for attr in (
        "process_name_to_controller_id",
        "controller_id_to_task",
        "task_to_processes",
    ):
        monkeypatch.setattr(f"ampel.run.server.task_manager.{attr}", {})

    async with AsyncClient(app=server.app, base_url="http://test") as client:
        yield client


@pytest.mark.asyncio
async def test_metrics(test_client):
    response = await test_client.get("/metrics")
    assert response.status_code == 200


@pytest.fixture
def db_collector(dev_context):
    c = AmpelDBCollector(dev_context.db)
    AmpelMetricsRegistry.register_collector(c)
    yield
    AmpelMetricsRegistry.deregister_collector(c)


@pytest.mark.asyncio
async def test_db_metrics(test_client, db_collector, dev_context):
    async def check_metric(name, value):
        response = await test_client.get("/metrics")
        assert response.status_code == 200
        for metric in text_fd_to_metric_families(StringIO(response.text)):
            if metric.name == name:
                assert len(metric.samples) == 1
                assert metric.samples[0].value == value
                break
        else:
            raise ValueError(f"metric {name} not collected")

    await check_metric("ampel_t2_docs_queued", 0)
    dev_context.db.get_collection("t2").insert_one({"code": DocumentCode.NEW})
    await check_metric("ampel_t2_docs_queued", 1)


@pytest.mark.parametrize(
    "section,proc",
    [
        (
            "t2",
            {
                "name": "DefaultT2Process",
                "version": 0,
                "active": True,
                "tier": 2,
                "schedule": ["every(5).minutes"],
                "processor": {
                    "unit": "T2Worker",
                    "config": {"doc_limit": 5000, "log_profile": "prod"},
                },
            },
        ),
        (
            "ops",
            {
                "name": "ExceptionPublisher",
                "version": 0,
                "active": True,
                "tier": None,
                "schedule": ["every(10).minutes"],
                "processor": {
                    "unit": "OpsProcessor",
                    "config": {"execute": {"unit": "AmpelExceptionPublisher"}},
                },
            },
        ),
    ],
)
@pytest.mark.asyncio
async def test_reload(
    test_client,
    dev_context,
    testing_config,
    monkeypatch,
    tmp_path,
    mocker,
    section,
    proc,
):
    config = yaml.safe_load(testing_config.open())
    config["process"][section][proc["name"]] = proc
    with open(tmp_path / "config.yml", "w") as f:
        yaml.dump(config, f)
    monkeypatch.setenv("AMPEL_CONFIG", str(tmp_path / "config.yml"))
    add_processes = mocker.patch("ampel.run.server.task_manager.add_processes")
    assert server.context is dev_context
    response = await test_client.post("/config/reload")
    assert response.status_code == 200
    assert server.context is not dev_context
    procs = add_processes.call_args.args[0]
    assert len(procs) == 1
    assert procs[0].processor.unit == proc["processor"]["unit"]


@pytest.mark.asyncio
async def test_processes(test_client):
    response = await test_client.get("/processes")
    assert response.status_code == 200
    processes = response.json()["processes"]
    assert len(processes) == 0


@pytest.mark.asyncio
async def test_processes_start(test_client):
    dict.__setitem__(
        server.context.config._config["process"]["t2"],
        "sleepy",
        {
            "name": "sleepy",
            "version": 0,
            "schedule": "every(30).seconds",
            "tier": 2,
            "isolate": True,
            "controller": {
                "unit": "DefaultProcessController",
                "config": {"mp_join": 2},
            },
            "processor": {"unit": "Sleepy"},
        },
    )
    try:
        response = await test_client.post("/processes/start", params={"tier": 2})
        assert response.status_code == 200
        assert (await test_client.get("/tasks")).json() == response.json()
        tasks = response.json()["tasks"]
        assert len(tasks) == 1
        assert tasks[0]["processes"] == ["sleepy"]

        assert len(tasks := list(server.task_manager.task_to_processes.keys())) == 1
        await asyncio.wait_for(tasks[0], 5)
        assert not (await test_client.get("/tasks")).json()["tasks"]
    finally:
        await server.task_manager.shutdown()


@pytest.mark.parametrize(
    "patches,should_raise",
    [
        (None, False),
        ({"processor.config": {"nonexistant_param": True}}, True),
        ({"active": False, "processor.config": {"nonexistant_param": True}}, False),
    ],
)
@pytest.fixture
def config_in_env(monkeypatch, tmp_path, dev_context, patches, should_raise):
    config = recursive_unfreeze(dev_context.config.get())
    config["process"]["t3"]["sleepy"] = {
        "name": "sleepy",
        "schedule": "every(30).seconds",
        "tier": 3,
        "isolate": True,
        "controller": {
            "unit": "DefaultProcessController",
            "config": {"mp_join": 2},
        },
        "processor": {"unit": "Sleepy"},
    }
    if patches:
        for k, v in patches.items():
            set_by_path(config["process"]["t3"]["sleepy"], k, v)
    with open(tmp_path / "config.yaml", "w") as f:
        yaml.dump(config, f)
    monkeypatch.setenv("AMPEL_CONFIG", str(tmp_path / "config.yaml"))


@pytest.mark.parametrize(
    "patches,should_raise",
    [
        ({}, False),
        ({"processor.config": {"nonexistant_param": True}}, True),
        ({"active": False, "processor.config": {"nonexistant_param": True}}, False),
    ],
)
@pytest.mark.asyncio
async def test_config_reload(
    test_client, monkeypatch, tmp_path, dev_context, patches, should_raise, mocker
):

    config = recursive_unfreeze(dev_context.config.get())
    config["process"]["t3"]["sleepy"] = {
        "name": "sleepy",
        "schedule": "every(30).seconds",
        "tier": 3,
        "version": 0,
        "isolate": True,
        "controller": {
            "unit": "DefaultProcessController",
            "config": {"mp_join": 2},
        },
        "processor": {"unit": "Sleepy"},
    }
    for k, v in patches.items():
        set_by_path(config["process"]["t3"]["sleepy"], k, v)
    with open(tmp_path / "config.yaml", "w") as f:
        yaml.dump(config, f)
    monkeypatch.setenv("AMPEL_CONFIG", str(tmp_path / "config.yaml"))

    try:
        remove = mocker.patch("ampel.run.server.task_manager.remove_processes")
        add = mocker.patch("ampel.run.server.task_manager.add_processes")
        response = await test_client.post("/config/reload")
        if should_raise:
            assert response.status_code == 500
        else:
            response.raise_for_status()
            assert await remove.called_once()
            assert len(remove.call_args[0][0]) == 0
            assert await add.called_once()
            assert len(add.call_args[0][0]) == (
                1 if config["process"]["t3"]["sleepy"].get("active", True) else 0
            )
    finally:
        await server.task_manager.shutdown()


@pytest.mark.asyncio
async def test_process_stop(test_client):
    dict.__setitem__(
        server.context.config._config["process"]["t2"],
        "sleepy",
        {
            "name": "sleepy",
            "schedule": "every(30).seconds",
            "version": 0,
            "tier": 2,
            "isolate": True,
            "processor": {"unit": "Sleepy"},
        },
    )
    try:
        response = await test_client.post("/processes/start", params={"tier": 2})
        assert response.status_code == 200
        assert (await test_client.get("/tasks")).json() == response.json()
        tasks = response.json()["tasks"]
        assert len(tasks) == 1
        assert tasks[0]["processes"] == ["sleepy"]

        assert len(tasks := list(server.task_manager.task_to_processes.keys())) == 1

        await asyncio.sleep(0.1)
        await test_client.post("/process/sleepy/stop")

        await asyncio.wait_for(tasks[0], 5)
        assert not (await test_client.get("/tasks")).json()["tasks"]
    finally:
        await server.task_manager.shutdown()


@pytest.mark.asyncio
async def test_event_query(test_client, mocker):
    m = mocker.patch("ampel.run.server.context.db")
    find = m.get_collection().find
    await test_client.get(
        "/events", params={"after": 7200, "process": "InfantSNSummary"}
    )
    assert isinstance(query := find.call_args.args[0], dict)
    assert isinstance(andlist := query["$and"], list)
    gtime = andlist[0]["_id"]["$gt"].generation_time
    assert 7200 < (datetime.now(gtime.tzinfo) - gtime).total_seconds() < 7230



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/utils/stock.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 29.04.2020
# Last Modified Date: 29.04.2020
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from pymongo import MongoClient
from pymongo.collection import Collection
from multiprocessing import Pool
from typing import Union, Optional, Set, Dict, Sequence
from ampel.types import ChannelId, StockId


def get_ids_using_find(
	col: Collection,
	channel: Union[ChannelId, Sequence[ChannelId]],
	batch_size=1000000
) -> Dict[ChannelId, Set[StockId]]:

	if isinstance(channel, (int, str)):
		return {
			channel: {
				el['_id'] for el in col \
					.find({'channel': channel}, {'_id': 1}) \
					.batch_size(batch_size)
			}
		}

	return {k: get_ids_using_find(col, k, batch_size)[k] for k in channel}


def get_ids_using_parallel_find(
	channel: Union[ChannelId, Sequence[ChannelId]],
	mongo_uri: Optional[str] = None,
	db_name='Ampel_data', col_name='stock', *,
	pool_size=None, batch_size=1000000
) -> Dict[ChannelId, Set[StockId]]:

	if isinstance(channel, (int, str)):
		return get_ids_using_find(
			MongoClient(mongo_uri) \
				.get_database(db_name) \
				.get_collection(col_name),
			channel, batch_size
		)

	pool = Pool(processes=pool_size)

	ret: Dict[ChannelId, Set[StockId]] = {k: set() for k in channel}

	results = [
		pool.apply_async(
			find_ids_worker,
			(k, mongo_uri, db_name, col_name, batch_size),
			callback = ret[k].update
		)
		for k in channel
	]

	for r in results:
		r.wait()

	pool.close()
	return ret


def find_ids_worker(
	channel: Union[ChannelId, Sequence[ChannelId]],
	mongo_uri: Optional[str] = None, db_name='Ampel_data',
	col_name='stock', batch_size=1000000
) -> Set[StockId]:

	mc = MongoClient(mongo_uri)
	db = mc.get_database(db_name)

	return {
		el['_id'] for el in db.get_collection(col_name) \
			.find({'channel': channel}, {'_id': 1}) \
			.batch_size(batch_size)
	}



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/model/time/TimeStringModel.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 29.09.2018
# Last Modified Date: 06.06.2020
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from typing import Literal
from datetime import datetime
from pydantic import BaseModel


class TimeStringModel(BaseModel):

	match_type: Literal['time_string']
	dateTimeStr: str
	dateTimeFormat: str

	def get_timestamp(self, **kwargs) -> float:

		return datetime \
			.strptime(self.dateTimeStr, self.dateTimeFormat) \
			.timestamp()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/cli/BufferCommand.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 25.03.2021
# Last Modified Date: 17.07.2021
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

import sys
from argparse import ArgumentParser
from typing import Sequence, Dict, Any, Optional, Union
from ampel.log.LogFlag import LogFlag
from ampel.core.AmpelContext import AmpelContext
from ampel.cli.utils import maybe_load_idmapper
from ampel.cli.ArgParserBuilder import ArgParserBuilder
from ampel.cli.AbsStockCommand import AbsStockCommand
from ampel.cli.AbsLoadCommand import AbsLoadCommand
from ampel.cli.AmpelArgumentParser import AmpelArgumentParser
from ampel.model.UnitModel import UnitModel
from ampel.t3.T3Processor import T3Processor

hlp = {
	'config': 'Path to an ampel config file (yaml/json)',
	# Optional
	'secrets': 'Path to a YAML secrets store in sops format',
	'out': 'Path to file where serialized views will be saved (printed to stdout otherwise)',
	'log-profile': 'One of: default, compact, headerless, verbose, debug',
	'id-mapper': 'Convert stock ids using the provided id mapper (ex: ZTFIdMapper)',
	'binary': 'Store buffers as compact BSON structures',
	'pretty': 'Show buffers using spacier prettyjson',
	'getch': 'Wait for a key to be pressed in terminal before printing next buffer (not compatible with binary option)',
}

class BufferCommand(AbsStockCommand, AbsLoadCommand):
	"""
	TODO: implement
	"""

	def __init__(self):
		self.parsers = {}

	# Mandatory implementation
	def get_parser(self, sub_op: Optional[str] = None) -> Union[ArgumentParser, AmpelArgumentParser]:

		if sub_op in self.parsers:
			return self.parsers[sub_op]

		sub_ops = ['show', 'save']
		if sub_op is None or sub_op not in sub_ops:
			return AmpelArgumentParser.build_choice_help(
				'buffer', sub_ops, hlp, description = 'Show or save ampel buffers.'
			)

		builder = ArgParserBuilder("buffer")
		hlp.update(self.get_select_args_help())
		hlp.update(self.get_load_args_help())
		builder.add_parsers(sub_ops, hlp)

		# Required args
		builder.add_arg("required", "config", type=str)
		builder.add_arg("save.required", "out", default=True)

		# Optional args
		builder.add_arg("optional", "secrets", default=None)
		builder.add_arg('optional', 'id-mapper')
		builder.add_arg('optional', 'debug', action="store_true")
		builder.add_arg("optional", "log-profile", default="default")
		builder.add_arg("save.optional", "binary", action="store_true")
		builder.add_arg("show.optional", "pretty", action="store_true")
		builder.add_arg("show.optional", "getch", action="store_true")

		# Selection args
		self.add_selection_args(builder)

		# Content args
		self.add_load_args(builder, 'Views content arguments')

		# Example
		builder.add_example("show", "-config ampel_conf.yaml -stock 85628462 -no-t0")
		builder.add_example("save", "-config ampel_conf.yaml -stock 85628462 -no-t0 -out ZTFabcdef.json")
		
		self.parsers.update(
			builder.get()
		)

		return self.parsers[sub_op]


	# Mandatory implementation
	def run(self, args: Dict[str, Any], unknown_args: Sequence[str], sub_op: Optional[str] = None) -> None:

		ctx: AmpelContext = self.get_context(args, unknown_args)
		maybe_load_idmapper(args)

		conf = {
			'fd': open(args["out"], "wb" if args.get('binary') else 'w') \
				if sub_op == "save" else sys.stdout,
			'id_mapper': args["id_mapper"],
			'verbose': sub_op == "save"
		}

		if args.get('pretty'):
			conf['pretty'] = True

		if not args.get('binary') and args.get('getch'):
			conf['getch'] = True

		t3p = T3Processor(
			context = ctx,
			process_name = "BufferCommand",
			base_log_flag = LogFlag.MANUAL_RUN,
			log_profile = 'console_debug' if args.get('debug') else 'console_info',
			update_journal = False,
			update_events = False,
			channel = args['channel'],
			select = self.build_select_model(args),
			load = self.build_load_model(args, codec_options=None),
			stage = UnitModel(
				unit="T3BufferBinaryExporter" if args.get('binary') \
					else "T3BufferTextExporter",
				config=conf
			)
		)

		t3p.run()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/demo/DemoReviewT3Unit.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 09.06.2020
# Last Modified Date: 17.12.2021
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from typing import Union, Generator
from ampel.types import UBson, T3Send
from ampel.abstract.AbsT3ReviewUnit import AbsT3ReviewUnit
from ampel.struct.JournalAttributes import JournalAttributes
from ampel.struct.UnitResult import UnitResult
from ampel.view.SnapView import SnapView
from ampel.view.T3Store import T3Store


class DemoReviewT3Unit(AbsT3ReviewUnit[SnapView]):

	parameter: int = 10

	def process(self,
		gen: Generator[SnapView, T3Send, None],
		t3s: T3Store
	) -> Union[UBson, UnitResult]:

		self.logger.info(f"DemoReviewT3Unit output (parameter={self.parameter}):")

		for v in gen:

			self.logger.info("id: " + str(v.id))
			gen.send(
				# This journal customization will be applied only to the current 'stock'
				JournalAttributes(tag="DemoT3SpecificTag", extra={'a': 1})
			)

		# This journal customization will be applied to all stocks
		return UnitResult(
			body = {'param': 'value'},
			code = 10,
			journal = JournalAttributes(tag="DemoReviewT3UnitTag")
		)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/cli/AbsLoadCommand.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 25.03.2021
# Last Modified Date: 29.03.2021
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from typing import Dict, Any, Optional
from bson.codec_options import CodecOptions
from ampel.cli.ArgParserBuilder import ArgParserBuilder
from ampel.cli.AbsCoreCommand import AbsCoreCommand
from ampel.model.UnitModel import UnitModel

# Maps args with load aliases
arg_doc_mapping = {
	'no_stock': "STOCK",
	'no_t0': "DATAPOINT",
	'no_t1': "COMPOUND",
	'no_t2': "T2DOC",
	'no_t3': "T3DOC"
}

class AbsLoadCommand(AbsCoreCommand, abstract=True):
	"""
	Base for commands that load ampel data using AbsT3Loader subclasses
	"""

	@staticmethod
	def get_load_args_help() -> Dict[str, str]:
		return {
			'latest': 'Include only latest state (*)',
			'no-stock': 'Exclude stock document from view',
			'no-plots': 'Exclude plots',
			'no-logs': 'Exclude logs'
		}

	def add_load_args(self, builder: ArgParserBuilder, group_description: str) -> None:

		# Content args
		builder.add_group('content', group_description)
		builder.add_arg('content', 'latest', action='store_true')
		builder.add_arg('content', 'no-stock', action='store_true')
		for el in (0, 1, 2, 3):
			builder.add_arg('content', f'no-t{el}', action='store_true', help=f"Exclude t{el} documents from view")
		builder.add_arg('content', 'no-plots', action='store_true')
		builder.add_arg('content', 'no-logs', action='store_true', help="Exclude logs")

		# Note
		builder.add_all_note("Latest state means... [adequate description in a few words]")


	def build_load_model(self, args: Dict[str, Any], codec_options: Optional[CodecOptions] = None) -> UnitModel:
		return UnitModel(
			unit = "T3LatestStateDataLoader" if args.get("latest") else "T3SimpleDataLoader",
			config = {
				'channel': args['channel'],
				'codec_options': codec_options,
				'directives': [
					v for k, v in arg_doc_mapping.items() if not args[k]
				]
			}
		)



from typing import Any
from ampel.abstract.AbsChannelTemplate import AbsChannelTemplate
from ampel.config.builder.FirstPassConfig import FirstPassConfig
from ampel.log.AmpelLogger import AmpelLogger
from ampel.model.ChannelModel import ChannelModel

class ChannelTemplate(AbsChannelTemplate):
    def get_processes(self, logger: AmpelLogger, first_pass_config: FirstPassConfig) -> list[dict[str, Any]]:
        return []
    

def test_AbsChannelTemplate(ampel_logger):

    template = ChannelTemplate(
        channel = "FOO",
        version = 0,
    )
    channel = template.get_channel(logger=ampel_logger)
    assert ChannelModel(**channel).__dict__ == channel
    


#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/util/distrib.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 13.03.2021
# Last Modified Date: 13.03.2021
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

import os, re
from typing import Generator, List, Optional
from pkg_resources import ( # type: ignore[attr-defined]
	get_distribution, AvailableDistributions,
	EggInfoDistribution, DistInfoDistribution
)


def get_dist_names(distrib_prefix: str = "ampel-") -> List[str]:
	"""
	Get all installed distributions whose names start with the provided prefix
	"""
	ret = [
		dist_name for dist_name in AvailableDistributions()
		if distrib_prefix in dist_name
	]

	if ret:
		ret.insert(0, ret.pop([i for i, el in enumerate(ret) if "interface" in el][0]))
		ret.insert(0, ret.pop([i for i, el in enumerate(ret) if "core" in el][0]))

	return ret


def get_files(dist_name: str, lookup_dir: Optional[str] = None, pattern: Optional[re.Pattern] = None) -> List[str]:
	"""
	Loads all known conf files of the provided distribution (name)
	"""

	if lookup_dir and not lookup_dir.endswith("/"):
		lookup_dir += "/"

	distrib = get_distribution(dist_name)

	# DistInfoDistribution: look for metadata RECORD (in <dist_name>.dist-info)
	if isinstance(distrib, DistInfoDistribution):
		# Example of the ouput of distrib.get_metadata_lines('RECORD'):
		# 'conf/ampel-ztf.conf,sha256=FZkChNKKpcMPTO4pwyKq4WS8FAbznuR7oL9rtNYS7U0,322',
		# 'ampel/model/ZTFLegacyChannelTemplate.py,sha256=zVtv4Iry3FloofSazIFc4h8l6hhV-wpIFbX3fOW2njA,2182',
		# 'ampel/model/__pycache__/ZTFLegacyChannelTemplate.cpython-38.pyc,,',
		if pth := next(
			(
				pth for el in distrib.get_metadata_lines('RECORD')
				if (pth := el.split(",")[0]).endswith(".pth")
			),
			None
		):
			return list(walk_pth_file(pth if os.path.isfile(pth) else distrib.get_resource_filename(__name__, pth), lookup_dir, pattern))
		else:
			return [
				fname for el in distrib.get_metadata_lines('RECORD')
				if _check_match((fname := el.split(",")[0]), lookup_dir, pattern)
			]

	elif isinstance(distrib, EggInfoDistribution):
		# Example of the ouput of distrib.get_metadata_lines('SOURCES.txt'):
		# 'setup.py',
		# 'conf/ampel-ztf.json',
		# 'ampel/model/ZTFLegacyChannelTemplate.py',
		return [
			el for el in distrib.get_metadata_lines('SOURCES.txt')
			if _check_match(el, lookup_dir, pattern)
		]

	else:
		raise ValueError(f"Unsupported distribution type: '{type(distrib)}'")


def _check_match(arg: str, lookup_dir: Optional[str], pattern: Optional[re.Pattern]) -> bool:
	if lookup_dir and not arg.startswith(lookup_dir):
		return False
	if pattern and not pattern.match(arg):
		return False
	return True


def walk_pth_file(pth: str, lookup_dir: Optional[str] = None, pattern: Optional[re.Pattern] = None) -> Generator[str, None, None]:
	with open(pth) as f:
		for root, dirs, files in os.walk(f"{f.read().strip()}/{lookup_dir}" if lookup_dir else f"{f.read().strip()}"):
			for fname in files:
				if pattern and not pattern.match(fname):
					continue
				yield os.path.join(root, fname)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/model/ingest/DualIngestDirective.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 20.05.2021
# Last Modified Date: 21.05.2021
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from typing import Optional, Literal, Dict
from ampel.types import ChannelId
from ampel.model.StrictModel import StrictModel
from ampel.model.ingest.IngestBody import IngestBody
from ampel.model.ingest.FilterModel import FilterModel


class DualIngestDirective(StrictModel):
	"""
    channel
    filter
    ingest
      new               <- stock was never previously associated with the current channel
        stock_t2
        point_t2        <- based on input dps list
        combine         <- same
          state_t2      <- based on dps list returned by combine
          point_t2      <- same
        mux             <- based on input dps list
          insert
            point_t2    <- based on <datapoints insert> result from muxer unit (first list)
          combine
            state_t2    <- based on <datapoints combine> result from muxer unit (second list)
            point_t2    <- same
      known             <- stock was previously associated with the current channel
        stock_t2
        point_t2        <- based on input dps list
        combine         <- same
          state_t2      <- based on dps list returned by combine
          point_t2      <- same
        mux             <- based on input dps list
          insert
            point_t2    <- based on <datapoints insert> result from muxer unit (first list)
          combine
            state_t2    <- based on <datapoints combine> result from muxer unit (second list)
            point_t2    <- same
	"""

	#: Channel for which to create documents
	channel: ChannelId

	#: Potientially filter input datapoints
	filter: Optional[FilterModel]

	ingest: Dict[Literal['new', 'known'], IngestBody]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/mongo/update/var/DBLoggingHandler.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 14.12.2017
# Last Modified Date: 29.03.2021
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

import struct, socket
from bson import ObjectId
from typing import List, Dict, Optional, Union, TYPE_CHECKING
from logging import LogRecord
from pymongo.errors import BulkWriteError
from ampel.base.AmpelBaseModel import AmpelBaseModel
from ampel.util.mappings import compare_dict_values
from ampel.util.collections import try_reduce
from ampel.log.AmpelLogger import AmpelLogger
from ampel.log.LightLogRecord import LightLogRecord
from ampel.log.AmpelLoggingError import AmpelLoggingError
from ampel.log.LogFlag import LogFlag
from ampel.log.utils import log_exception, report_exception

if TYPE_CHECKING:
	from ampel.core.AmpelDB import AmpelDB


# http://isthe.com/chongo/tech/comp/fnv/index.html#FNV-1a
def _fnv_1a_24(data, _ord=lambda x: x):
	"""FNV-1a 24 bit hash"""
	# http://www.isthe.com/chongo/tech/comp/fnv/index.html#xor-fold
	# Start with FNV-1a 32 bit.
	hash_size = 2 ** 32
	fnv_32_prime = 16777619
	fnv_1a_hash = 2166136261  # 32-bit FNV-1 offset basis
	for elt in data:
		fnv_1a_hash = fnv_1a_hash ^ _ord(elt)
		fnv_1a_hash = (fnv_1a_hash * fnv_32_prime) % hash_size

	# xor-fold the result to 24 bit.
	return (fnv_1a_hash >> 24) ^ (fnv_1a_hash & 0xffffff)


def _machine_bytes():
	"""Get the machine portion of an ObjectId.
	"""
	# gethostname() returns a unicode string in python 3.x
	# We only need 3 bytes, and _fnv_1a_24 returns a 24 bit integer.
	# Remove the padding byte.
	return struct.pack("<I", _fnv_1a_24(socket.gethostname().encode()))[:3]


class DBLoggingHandler(AmpelBaseModel):
	""" Saves log events into mongo database """

	__slots__ = "prev_record", "fields_check", "log_dicts", "oid_middle", "warn_lvl"

	level: int
	col_name: str = "logs"
	aggregate_interval: float = 1
	expand_extra: bool = True
	flush_len: int = 1000
	auto_flush: bool = False


	def __init__(self, ampel_db: 'AmpelDB', run_id: int, **kwargs) -> None:
		"""
		:param col_name: name of db collection to use (default: 'logs' in database Ampel_var)
		:param aggregate_interval: logs with similar attributes (log level, possibly tranId & channels) \
		are aggregated in one document instead of being split into several documents (spares some index RAM). \
		*aggregate_interval* is the max interval of time in seconds during which log aggregation takes place. \
		Beyond this value, a new log document is created no matter what. This parameter thus impacts logging time granularity.
		:param flush_len: How many log documents should be kept in memory before attempting a database bulk_write operation.
		"""

		AmpelBaseModel.__init__(self, **kwargs)
		self._ampel_db = ampel_db
		self.run_id = run_id

		self.log_dicts: List[Dict] = []
		self.prev_record: Optional[Union[LightLogRecord, LogRecord]] = None
		self.fields_check = ['extra', 'stock', 'channel']
		self.warn_lvl = LogFlag.WARNING

		# Get reference to pymongo collection
		self.col = ampel_db.get_collection(self.col_name)

		# ObjectID middle: 3 bytes machine + 2 bytes encoding the last 4 digits of run_id (unique)
		# NB: pid is not always unique if running in a jail or container
		self.oid_middle = _machine_bytes() + int(str(self.run_id)[-4:]).to_bytes(2, 'big')


	def handle(self, record: Union[LightLogRecord, LogRecord]) -> None:
		""" :raises AmpelLoggingError: on error """

		rd = record.__dict__
		prev_record = self.prev_record

		try:

			# Same flag, date (+- 1 sec), tran_id and chans
			if (
				prev_record and
				(record.name is None or record.name == prev_record.name) and
				record.levelno <= prev_record.levelno and
				record.created - prev_record.created < self.aggregate_interval and
				compare_dict_values(rd, prev_record.__dict__, self.fields_check)
			):

				prev_dict = self.log_dicts[-1]

				if 'm' not in prev_dict:
					prev_dict['m'] = "None log entry with identical fields repeated twice"
					return

				if isinstance(prev_dict['m'], list):
					prev_dict['m'].append(record.msg)
				else:
					prev_dict['m'] = [prev_dict['m'], record.msg]

			else:

				if self.auto_flush and len(self.log_dicts) > self.flush_len:
					self.flush()

				# Treat SHOUT msg as INFO msg (and try again to concatenate)
				if record.levelno == LogFlag.SHOUT: # type: ignore
					if isinstance(record, LightLogRecord):
						new_rec = LightLogRecord(name=0, msg=None, levelno=0)
					else:
						new_rec = LogRecord(name=None, pathname=None, level=None, # type: ignore
							lineno=None, exc_info=None, msg=None, args=None) # type: ignore
					for k, v in record.__dict__.items():
						new_rec.__dict__[k] = v
					new_rec.levelno = LogFlag.INFO
					self.handle(new_rec)
					return

				# Generate object id with log record.created as current time
				with ObjectId._inc_lock:
					oid = struct.pack(">i", int(record.created)) + \
						self.oid_middle + struct.pack(">i", ObjectId._inc)[1:4]
					ObjectId._inc = (ObjectId._inc + 1) % 0xFFFFFF # limit result to 32bits

				if 'extra' in rd:
					if self.expand_extra:
						for k in ('_id', 'r', 'f'):
							if k in rd['extra']:
								del rd['extra'][k]
						ldict = {
							'_id': ObjectId(oid=oid),
							'r': self.run_id,
							'f': record.levelno,
							**rd['extra']
						}
					else:
						ldict = {
							'_id': ObjectId(oid=oid),
							'r': self.run_id,
							'f': record.levelno,
							'x': rd['extra']
						}
				else:
					ldict = {
						'_id': ObjectId(oid=oid),
						'r': self.run_id,
						'f': record.levelno
					}

				if 'stock' in rd:
					ldict['s'] = rd['stock']

				if 'channel' in rd:
					ldict['c'] = try_reduce(rd['channel'])

				if record.levelno > self.warn_lvl:
					ldict['file'] = record.filename
					ldict['line'] = record.lineno

				if record.msg:
					ldict['m'] = record.msg

				self.log_dicts.append(ldict)
				self.prev_record = record

		except Exception as e:

			logger = AmpelLogger.get_logger()

			# Print log stack using std logging
			log_exception(logger, e, msg="Exception:")

			try:
				# This will fail as well if we have DB connectivity issues
				report_exception(self._ampel_db, logger, e)
			except Exception as ee:
				log_exception(
					logger, ee, last=True,
					msg="Could not update troubles collection as well (DB offline?)"
				)

			logger.error("DB log flushing error, un-flushed (json) logs below.")
			logger.error("*" * 52)

			from ampel.util.pretty import prettyjson
			for d in self.log_dicts:
				logger.error(prettyjson(d))
			logger.error("#" * 52)

			raise AmpelLoggingError from None


	def get_run_id(self) -> int:
		return self.run_id


	def check_flush(self) -> None:
		if len(self.log_dicts) > self.flush_len:
			self.flush()


	def break_aggregation(self) -> None:
		self.prev_record = None


	def flush(self) -> None:
		""" :raises AmpelLoggingError: on error """

		# No log entries
		if not self.log_dicts:
			return

		# save referenced log entries locally
		dicts = self.log_dicts

		# clear referenced log entries
		self.log_dicts = []
		self.prev_record = None

		try:
			# pymongo drops the GIL while sending and receiving data over the network
			self.col.insert_many(dicts, ordered=False)

		except BulkWriteError as bwe:
			if self.handle_bulk_write_error(bwe):
				raise AmpelLoggingError from None

		except Exception as e:
			from ampel.log.LoggingErrorReporter import LoggingErrorReporter
			LoggingErrorReporter.report(self, e)
			# If we can no longer keep track of what Ampel is doing,
			# better raise Exception to stop processing
			raise AmpelLoggingError from None


	def handle_bulk_write_error(self, bwe: BulkWriteError) -> bool:
		"""
		:returns: true if error could not be handled properly
		"""
		# We try here to handle duplicate key errors which can occur in rare scenarios.
		# We loop through each element in the returned writeErrors and
		# see for the E11000 cases if - in the DB - the runId of the log record
		# with conflicting ObjectId is the same as the one from the current instance.
		# If the runId are the same, then the conflicting log entry originates from
		# the current instance and we trust it as being the same than the one in the buffer
		# Further equality checks (msg content) could be done but we don't do them for now.

		try:

			from ampel.log.utils import get_tier_from_log_flags, convert_dollars, log_exception
			raise_exc = False

			for err_dict in bwe.details.get('writeErrors', []):

				# This log entry was already inserted into DB
				# 'code': 11000, 'errmsg': 'E11000 duplicate key error collection: ...
				if err_dict.get("code") == 11000:

					# Fetch corresponding log entry from DB
					db_rec = next(self.col.find({'_id': err_dict['op']['_id']}))

					# if runIds are equal, just print out a feedback
					if db_rec['run'] == self.run_id:
						print("Disregardable E11000: %s" % err_dict['op'])

					# Otherwise print error and create trouble doc
					else:

						raise_exc = True

						print("CRITICAL: OID collision occured between two different log entries")
						print("Current process:")
						print(err_dict['op'])
						print("In DB:")
						print(db_rec)

						self._ampel_db.get_collection('troubles').insert_one(
							{
								'tier': get_tier_from_log_flags(db_rec['f']),
								'location': 'DBLoggingHandler',
								'msg': "OID collision occured between two different log entries",
								'ownLogEntry': str(err_dict['op']),
								'existingLogEntry': str(db_rec)
							}
						)
				else:

					raise_exc = True
					print("writeError dict entry: %s" % err_dict)

					self._ampel_db.get_collection('troubles').insert_one(
						{
							'location': 'DBLoggingHandler',
							'msg': "non-E11000 writeError occured",
							'errorDict': convert_dollars(err_dict)
						}
					)

			return raise_exc

		# bad day
		except Exception:

			try:
				from traceback import format_exc
				self._ampel_db.get_collection('troubles').insert_one(
					{
						'location': 'DBLoggingHandler',
						'msg': "Exception occured in handle_bulk_write_error",
						'exception': format_exc().replace("\"", "'").split("\n")
					}
				)
			except Exception as another_exc:
				from ampel.log.AmpelLogger import AmpelLogger
				log_exception(
					AmpelLogger.get_logger(),
					another_exc
				)

			from ampel.log.LoggingErrorReporter import LoggingErrorReporter
			LoggingErrorReporter.report(self, bwe, bwe.details)

			return True



import ast
from typing import Any, Optional


class ExpressionParser(ast.NodeVisitor):
    @classmethod
    def evaluate(cls, expression: str, context: dict) -> str:
        """Evaluate Argo-style template expressions"""
        tree = ast.parse(expression, mode="eval")
        self = cls(context)
        self.visit(tree)
        if self._value is None:
            raise ValueError(f"unable to resolve {expression}")
        return self._value

    def __init__(self, context: dict):
        self._context = context
        self._value: Optional[str] = None

    def visit_Attribute(self, node: ast.Attribute) -> None:
        path = [node.attr]
        parent = node.value
        while isinstance(parent, ast.Attribute):
            path.insert(0, parent.attr)
            parent = parent.value
        if isinstance(parent, ast.Name):
            path.insert(0, parent.id)
        context: Any = self._context
        while path:
            try:
                context = context[path.pop(0)]
            except KeyError:
                return
        if isinstance(context, str):
            self._value = context
        elif hasattr(context, "value"):
            self._value = context.value()




#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/model/time/QueryTimeModel.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 10.12.2019
# Last Modified Date: 06.06.2020
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from pydantic import Field
from typing import Union, Optional, Dict, Any
from ampel.model.StrictModel import StrictModel


class QueryTimeModel(StrictModel):
	"""
	Standardized parameter for the class QueryMatchStock
	"""
	before: Optional[Union[int, float]] = Field(None, alias='$lt')
	after: Optional[Union[int, float]] = Field(None, alias='$gt')

	def __bool__(self) -> bool:
		return self.before is not None or self.after is not None

	def dict(self, **kwargs) -> Dict[str, Any]:
		"""
		Example:
		{
			'$gt': 1575000000.003819,
			'$lt': 1575966106.003819
		}
		"""
		return super().dict(
			**{**kwargs, "by_alias": True, "exclude_none": True} # type: ignore
		)
